import unittest

from django.template import Library

import django_components as dc

from .django_test_setup import setup_test_config

setup_test_config()


class MockComponent(dc.Component):
    pass


class MockComponent2(dc.Component):
    pass


class MockComponentView(dc.Component):
    def get(self, request, *args, **kwargs):
        pass


class ComponentRegistryTest(unittest.TestCase):
    def setUp(self):
        self.registry = dc.ComponentRegistry()

    def test_register_class_decorator(self):
        @dc.register("decorated_component")
        class TestComponent(dc.Component):
            pass

        self.assertEqual(dc.registry.get("decorated_component"), TestComponent)

    def test_register_class_decorator_custom_registry(self):
        dc.registry.clear()

        my_lib = Library()
        my_reg = dc.ComponentRegistry(library=my_lib)

        self.assertDictEqual(my_reg.all(), {})
        self.assertDictEqual(dc.registry.all(), {})

        @dc.register("decorated_component", registry=my_reg)
        class TestComponent(dc.Component):
            pass

        self.assertDictEqual(my_reg.all(), {"decorated_component": TestComponent})
        self.assertDictEqual(dc.registry.all(), {})

    def test_simple_register(self):
        self.registry.register(name="testcomponent", component=MockComponent)
        self.assertEqual(self.registry.all(), {"testcomponent": MockComponent})

    def test_register_two_components(self):
        self.registry.register(name="testcomponent", component=MockComponent)
        self.registry.register(name="testcomponent2", component=MockComponent)
        self.assertEqual(
            self.registry.all(),
            {
                "testcomponent": MockComponent,
                "testcomponent2": MockComponent,
            },
        )

    def test_unregisters_only_unused_tags(self):
        self.assertDictEqual(self.registry._tags, {})
        self.assertNotIn("component", self.registry.library.tags)
        self.assertNotIn("#component", self.registry.library.tags)

        # Register two components that use the same tag
        self.registry.register(name="testcomponent", component=MockComponent)
        self.registry.register(name="testcomponent2", component=MockComponent)

        self.assertDictEqual(self.registry._tags, {
            "component": {"testcomponent", "testcomponent2"},
            "#component": {"testcomponent", "testcomponent2"},
        })

        self.assertIn("component", self.registry.library.tags)
        self.assertIn("#component", self.registry.library.tags)

        # Unregister only one of the components. The tags should remain
        self.registry.unregister(name="testcomponent")

        self.assertDictEqual(self.registry._tags, {
            "component": {"testcomponent2"},
            "#component": {"testcomponent2"},
        })

        self.assertIn("component", self.registry.library.tags)
        self.assertIn("#component", self.registry.library.tags)

        # Unregister the second components. The tags should be removed
        self.registry.unregister(name="testcomponent2")

        self.assertDictEqual(self.registry._tags, {})
        self.assertNotIn("component", self.registry.library.tags)
        self.assertNotIn("#component", self.registry.library.tags)

    def test_prevent_registering_different_components_with_the_same_name(self):
        self.registry.register(name="testcomponent", component=MockComponent)
        with self.assertRaises(dc.AlreadyRegistered):
            self.registry.register(name="testcomponent", component=MockComponent2)

    def test_allow_duplicated_registration_of_the_same_component(self):
        try:
            self.registry.register(name="testcomponent", component=MockComponentView)
            self.registry.register(name="testcomponent", component=MockComponentView)
        except dc.AlreadyRegistered:
            self.fail("Should not raise AlreadyRegistered")

    def test_simple_unregister(self):
        self.registry.register(name="testcomponent", component=MockComponent)
        self.registry.unregister(name="testcomponent")
        self.assertEqual(self.registry.all(), {})

    def test_raises_on_failed_unregister(self):
        with self.assertRaises(dc.NotRegistered):
            self.registry.unregister(name="testcomponent")



"""
Tests focusing on the Component class.
For tests focusing on the `component` tag, see `test_templatetags_component.py`
"""

from typing import Dict

from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponse
from django.template import Context, Template, TemplateSyntaxError

import django_components as dc
from django_components.slots import SlotRef

from .django_test_setup import setup_test_config
from .testutils import BaseTestCase, parametrize_context_behavior

setup_test_config({"autodiscover": False})


class ComponentTest(BaseTestCase):
    class ParentComponent(dc.Component):
        template: dc.django_html = """
            {% load component_tags %}
            <div>
                <h1>Parent content</h1>
                {% component name="variable_display" shadowing_variable='override' new_variable='unique_val' %}
                {% endcomponent %}
            </div>
            <div>
                {% slot 'content' %}
                    <h2>Slot content</h2>
                    {% component name="variable_display" shadowing_variable='slot_default_override' new_variable='slot_default_unique' %}
                    {% endcomponent %}
                {% endslot %}
            </div>
        """  # noqa

        def get_context_data(self):
            return {"shadowing_variable": "NOT SHADOWED"}

    class VariableDisplay(dc.Component):
        template: dc.django_html = """
            {% load component_tags %}
            <h1>Shadowing variable = {{ shadowing_variable }}</h1>
            <h1>Uniquely named variable = {{ unique_variable }}</h1>
        """

        def get_context_data(self, shadowing_variable=None, new_variable=None):
            context = {}
            if shadowing_variable is not None:
                context["shadowing_variable"] = shadowing_variable
            if new_variable is not None:
                context["unique_variable"] = new_variable
            return context

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        dc.registry.register(name="parent_component", component=cls.ParentComponent)
        dc.registry.register(name="variable_display", component=cls.VariableDisplay)

    @parametrize_context_behavior(["django", "isolated"])
    def test_empty_component(self):
        class EmptyComponent(dc.Component):
            pass

        with self.assertRaises(ImproperlyConfigured):
            EmptyComponent("empty_component").get_template(Context({}))

    @parametrize_context_behavior(["django", "isolated"])
    def test_template_string_static_inlined(self):
        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                Variable: <strong>{{ variable }}</strong>
            """

            def get_context_data(self, variable=None):
                return {
                    "variable": variable,
                }

            class Media:
                css = "style.css"
                js = "script.js"

        rendered = SimpleComponent.render(kwargs={"variable": "test"})
        self.assertHTMLEqual(
            rendered,
            """
            Variable: <strong>test</strong>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_template_string_dynamic(self):
        class SimpleComponent(dc.Component):
            def get_template_string(self, context):
                content: dc.django_html = """
                    Variable: <strong>{{ variable }}</strong>
                """
                return content

            def get_context_data(self, variable=None):
                return {
                    "variable": variable,
                }

            class Media:
                css = "style.css"
                js = "script.js"

        rendered = SimpleComponent.render(kwargs={"variable": "test"})
        self.assertHTMLEqual(
            rendered,
            """
            Variable: <strong>test</strong>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_template_name_static(self):
        class SimpleComponent(dc.Component):
            template_name = "simple_template.html"

            def get_context_data(self, variable=None):
                return {
                    "variable": variable,
                }

            class Media:
                css = "style.css"
                js = "script.js"

        rendered = SimpleComponent.render(kwargs={"variable": "test"})
        self.assertHTMLEqual(
            rendered,
            """
            Variable: <strong>test</strong>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_template_name_dynamic(self):
        class SvgComponent(dc.Component):
            def get_context_data(self, name, css_class="", title="", **attrs):
                return {
                    "name": name,
                    "css_class": css_class,
                    "title": title,
                    **attrs,
                }

            def get_template_name(self, context):
                return f"dynamic_{context['name']}.svg"

        self.assertHTMLEqual(
            SvgComponent.render(kwargs={"name": "svg1"}),
            """
            <svg>Dynamic1</svg>
            """,
        )
        self.assertHTMLEqual(
            SvgComponent.render(kwargs={"name": "svg2"}),
            """
            <svg>Dynamic2</svg>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_allows_to_override_get_template(self):
        class TestComponent(dc.Component):
            def get_context_data(self, variable, **attrs):
                return {
                    "variable": variable,
                }

            def get_template(self, context):
                template_str = "Variable: <strong>{{ variable }}</strong>"
                return Template(template_str)

        rendered = TestComponent.render(kwargs={"variable": "test"})
        self.assertHTMLEqual(
            rendered,
            """
            Variable: <strong>test</strong>
            """,
        )


class ComponentRenderTest(BaseTestCase):
    @parametrize_context_behavior(["django", "isolated"])
    def test_render_minimal(self):
        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                the_arg2: {{ the_arg2 }}
                args: {{ args|safe }}
                the_kwarg: {{ the_kwarg }}
                kwargs: {{ kwargs|safe }}
                ---
                from_context: {{ from_context }}
                ---
                slot_second: {% slot "second" default %}
                    SLOT_SECOND_DEFAULT
                {% endslot %}
            """

            def get_context_data(self, the_arg2=None, *args, the_kwarg=None, **kwargs):
                return {
                    "the_arg2": the_arg2,
                    "the_kwarg": the_kwarg,
                    "args": args,
                    "kwargs": kwargs,
                }

        rendered = SimpleComponent.render()
        self.assertHTMLEqual(
            rendered,
            """
            the_arg2: None
            args: ()
            the_kwarg: None
            kwargs: {}
            ---
            from_context:
            ---
            slot_second: SLOT_SECOND_DEFAULT
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_render_full(self):
        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                the_arg: {{ the_arg }}
                the_arg2: {{ the_arg2 }}
                args: {{ args|safe }}
                the_kwarg: {{ the_kwarg }}
                kwargs: {{ kwargs|safe }}
                ---
                from_context: {{ from_context }}
                ---
                slot_first: {% slot "first" required %}
                {% endslot %}
                ---
                slot_second: {% slot "second" default %}
                    SLOT_SECOND_DEFAULT
                {% endslot %}
            """

            def get_context_data(self, the_arg, the_arg2=None, *args, the_kwarg, **kwargs):
                return {
                    "the_arg": the_arg,
                    "the_arg2": the_arg2,
                    "the_kwarg": the_kwarg,
                    "args": args,
                    "kwargs": kwargs,
                }

        rendered = SimpleComponent.render(
            context={"from_context": 98},
            args=["one", "two", "three"],
            kwargs={"the_kwarg": "test", "kw2": "ooo"},
            slots={"first": "FIRST_SLOT"},
        )
        self.assertHTMLEqual(
            rendered,
            """
            the_arg: one
            the_arg2: two
            args: ('three',)
            the_kwarg: test
            kwargs: {'kw2': 'ooo'}
            ---
            from_context: 98
            ---
            slot_first: FIRST_SLOT
            ---
            slot_second: SLOT_SECOND_DEFAULT
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_render_to_response_full(self):
        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                the_arg: {{ the_arg }}
                the_arg2: {{ the_arg2 }}
                args: {{ args|safe }}
                the_kwarg: {{ the_kwarg }}
                kwargs: {{ kwargs|safe }}
                ---
                from_context: {{ from_context }}
                ---
                slot_first: {% slot "first" required %}
                {% endslot %}
                ---
                slot_second: {% slot "second" default %}
                    SLOT_SECOND_DEFAULT
                {% endslot %}
            """

            def get_context_data(self, the_arg, the_arg2=None, *args, the_kwarg, **kwargs):
                return {
                    "the_arg": the_arg,
                    "the_arg2": the_arg2,
                    "the_kwarg": the_kwarg,
                    "args": args,
                    "kwargs": kwargs,
                }

        rendered = SimpleComponent.render_to_response(
            context={"from_context": 98},
            args=["one", "two", "three"],
            kwargs={"the_kwarg": "test", "kw2": "ooo"},
            slots={"first": "FIRST_SLOT"},
        )
        self.assertIsInstance(rendered, HttpResponse)

        self.assertHTMLEqual(
            rendered.content.decode(),
            """
            the_arg: one
            the_arg2: two
            args: ('three',)
            the_kwarg: test
            kwargs: {'kw2': 'ooo'}
            ---
            from_context: 98
            ---
            slot_first: FIRST_SLOT
            ---
            slot_second: SLOT_SECOND_DEFAULT
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_render_to_response_change_response_class(self):
        class MyResponse:
            def __init__(self, content: str) -> None:
                self.content = bytes(content, "utf-8")

        class SimpleComponent(dc.Component):
            response_class = MyResponse
            template: dc.django_html = "HELLO"

        rendered = SimpleComponent.render_to_response()
        self.assertIsInstance(rendered, MyResponse)

        self.assertHTMLEqual(
            rendered.content.decode(),
            "HELLO",
        )

    @parametrize_context_behavior([("django", False), ("isolated", True)])
    def test_render_slot_as_func(self, context_behavior_data):
        is_isolated = context_behavior_data

        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                {% slot "first" required data1="abc" data2:hello="world" data2:one=123 %}
                    SLOT_DEFAULT
                {% endslot %}
            """

            def get_context_data(self, the_arg, the_kwarg=None, **kwargs):
                return {
                    "the_arg": the_arg,
                    "the_kwarg": the_kwarg,
                    "kwargs": kwargs,
                }

        def first_slot(ctx: Context, slot_data: Dict, slot_ref: SlotRef):
            self.assertIsInstance(ctx, Context)
            # NOTE: Since the slot has access to the Context object, it should behave
            # the same way as it does in templates - when in "isolated" mode, then the
            # slot fill has access only to the "root" context, but not to the data of
            # get_context_data() of SimpleComponent.
            if is_isolated:
                self.assertEqual(ctx.get("the_arg"), None)
                self.assertEqual(ctx.get("the_kwarg"), None)
                self.assertEqual(ctx.get("kwargs"), None)
                self.assertEqual(ctx.get("abc"), None)
            else:
                self.assertEqual(ctx["the_arg"], "1")
                self.assertEqual(ctx["the_kwarg"], 3)
                self.assertEqual(ctx["kwargs"], {})
                self.assertEqual(ctx["abc"], "def")

            slot_data_expected = {
                "data1": "abc",
                "data2": {"hello": "world", "one": 123},
            }
            self.assertDictEqual(slot_data_expected, slot_data)

            self.assertIsInstance(slot_ref, SlotRef)
            self.assertEqual("SLOT_DEFAULT", str(slot_ref).strip())

            return f"FROM_INSIDE_FIRST_SLOT | {slot_ref}"

        rendered = SimpleComponent.render(
            context={"abc": "def"},
            args=["1"],
            kwargs={"the_kwarg": 3},
            slots={"first": first_slot},
        )
        self.assertHTMLEqual(
            rendered,
            "FROM_INSIDE_FIRST_SLOT | SLOT_DEFAULT",
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_render_raises_on_missing_slot(self):
        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                {% slot "first" required %}
                {% endslot %}
            """

        with self.assertRaises(TemplateSyntaxError):
            SimpleComponent.render()

        SimpleComponent.render(
            slots={"first": "FIRST_SLOT"},
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_render_with_include(self):
        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                {% include 'slotted_template.html' %}
            """

        rendered = SimpleComponent.render()
        self.assertHTMLEqual(
            rendered,
            """
            <custom-template>
                <header>Default header</header>
                <main>Default main</main>
                <footer>Default footer</footer>
            </custom-template>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_render_with_extends(self):
        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                {% extends 'block.html' %}
                {% block body %}
                    OVERRIDEN
                {% endblock %}
            """

        rendered = SimpleComponent.render()
        self.assertHTMLEqual(
            rendered,
            """
            <!DOCTYPE html>
            <html lang="en">
            <body>
                <main role="main">
                <div class='container main-container'>
                    OVERRIDEN
                </div>
                </main>
            </body>
            </html>

            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_render_can_access_instance(self):
        class TestComponent(dc.Component):
            template = "Variable: <strong>{{ id }}</strong>"

            def get_context_data(self, **attrs):
                return {
                    "id": self.component_id,
                }

        rendered = TestComponent(component_id="123").render()
        self.assertHTMLEqual(
            rendered,
            "Variable: <strong>123</strong>",
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_render_to_response_can_access_instance(self):
        class TestComponent(dc.Component):
            template = "Variable: <strong>{{ id }}</strong>"

            def get_context_data(self, **attrs):
                return {
                    "id": self.component_id,
                }

        rendered_resp = TestComponent(component_id="123").render_to_response()
        self.assertHTMLEqual(
            rendered_resp.content.decode("utf-8"),
            "Variable: <strong>123</strong>",
        )



import os
import tempfile
from io import StringIO
from shutil import rmtree

from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase

from .django_test_setup import setup_test_config

setup_test_config()


class CreateComponentCommandTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.temp_dir = tempfile.mkdtemp()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        rmtree(cls.temp_dir)

    def test_default_file_names(self):
        component_name = "defaultcomponent"
        call_command("startcomponent", component_name, "--path", self.temp_dir)

        expected_files = [
            os.path.join(self.temp_dir, component_name, "script.js"),
            os.path.join(self.temp_dir, component_name, "style.css"),
            os.path.join(self.temp_dir, component_name, "template.html"),
        ]
        for file_path in expected_files:
            self.assertTrue(os.path.exists(file_path))

    def test_nondefault_creation(self):
        component_name = "testcomponent"
        call_command(
            "startcomponent",
            component_name,
            "--path",
            self.temp_dir,
            "--js",
            "test.js",
            "--css",
            "test.css",
            "--template",
            "test.html",
        )

        expected_files = [
            os.path.join(self.temp_dir, component_name, "test.js"),
            os.path.join(self.temp_dir, component_name, "test.css"),
            os.path.join(self.temp_dir, component_name, "test.html"),
            os.path.join(self.temp_dir, component_name, f"{component_name}.py"),
        ]

        for file_path in expected_files:
            self.assertTrue(os.path.exists(file_path), f"File {file_path} was not created")

    def test_dry_run(self):
        component_name = "dryruncomponent"
        call_command(
            "startcomponent",
            component_name,
            "--path",
            self.temp_dir,
            "--dry-run",
        )

        component_path = os.path.join(self.temp_dir, component_name)
        self.assertFalse(os.path.exists(component_path))

    def test_force_overwrite(self):
        component_name = "existingcomponent"
        component_path = os.path.join(self.temp_dir, component_name)
        os.makedirs(component_path)

        with open(os.path.join(component_path, f"{component_name}.py"), "w") as f:
            f.write("hello world")

        call_command(
            "startcomponent",
            component_name,
            "--path",
            self.temp_dir,
            "--force",
        )

        with open(os.path.join(component_path, f"{component_name}.py"), "r") as f:
            self.assertNotIn("hello world", f.read())

    def test_error_existing_component_no_force(self):
        component_name = "existingcomponent_2"
        component_path = os.path.join(self.temp_dir, component_name)
        os.makedirs(component_path)

        with self.assertRaises(CommandError):
            call_command("startcomponent", component_name, "--path", self.temp_dir)

    def test_verbose_output(self):
        component_name = "verbosecomponent"
        out = StringIO()
        call_command(
            "startcomponent",
            component_name,
            "--path",
            self.temp_dir,
            "--verbose",
            stdout=out,
        )
        output = out.getvalue()
        self.assertIn("component at", output)



from pathlib import Path

from django.contrib.staticfiles.management.commands.collectstatic import Command
from django.test import SimpleTestCase, override_settings

from .django_test_setup import *  # NOQA


# This subclass allows us to call the `collectstatic` command from within Python.
# We call the `collect` method, which returns info about what files were collected.
#
# The methods below are overriden to ensure we don't make any filesystem changes
# (copy/delete), as the original command copies files. Thus we can safely test that
# our `safer_staticfiles` app works as intended.
class MockCollectstaticCommand(Command):
    # NOTE: We do not expect this to be called
    def clear_dir(self, path):
        raise NotImplementedError()

    # NOTE: We do not expect this to be called
    def link_file(self, path, prefixed_path, source_storage):
        raise NotImplementedError()

    def copy_file(self, path, prefixed_path, source_storage):
        # Skip this file if it was already copied earlier
        if prefixed_path in self.copied_files:
            return self.log("Skipping '%s' (already copied earlier)" % path)
        # Delete the target file if needed or break
        if not self.delete_file(path, prefixed_path, source_storage):
            return
        # The full path of the source file
        source_path = source_storage.path(path)
        # Finally start copying
        if self.dry_run:
            self.log("Pretending to copy '%s'" % source_path, level=1)
        else:
            self.log("Copying '%s'" % source_path, level=2)
            # ############# OUR CHANGE ##############
            # with source_storage.open(path) as source_file:
            #     self.storage.save(prefixed_path, source_file)
            # ############# OUR CHANGE ##############
        self.copied_files.append(prefixed_path)


def do_collect():
    cmd = MockCollectstaticCommand()
    cmd.set_options(
        interactive=False,
        verbosity=1,
        link=False,
        clear=False,
        dry_run=False,
        ignore_patterns=[],
        use_default_ignore_patterns=True,
        post_process=True,
    )
    collected = cmd.collect()
    return collected


common_settings = {
    "STATIC_URL": "static/",
    "STATICFILES_DIRS": [Path(__file__).resolve().parent / "components"],
    "STATIC_ROOT": "staticfiles",
    "ROOT_URLCONF": __name__,
    "SECRET_KEY": "secret",
}


# Check that .py and .html files are INCLUDED with the original staticfiles app
@override_settings(
    **common_settings,
    INSTALLED_APPS=("django_components", "django.contrib.staticfiles"),
)
class OrigStaticFileTests(SimpleTestCase):
    def test_python_and_html_included(self):
        collected = do_collect()

        self.assertIn("safer_staticfiles/safer_staticfiles.css", collected["modified"])
        self.assertIn("safer_staticfiles/safer_staticfiles.js", collected["modified"])
        self.assertIn("safer_staticfiles/safer_staticfiles.html", collected["modified"])
        self.assertIn("safer_staticfiles/safer_staticfiles.py", collected["modified"])

        self.assertListEqual(collected["unmodified"], [])
        self.assertListEqual(collected["post_processed"], [])


# Check that .py and .html files are OMITTED from our version of staticfiles app
@override_settings(
    **common_settings,
    INSTALLED_APPS=("django_components", "django_components.safer_staticfiles"),
)
class SaferStaticFileTests(SimpleTestCase):
    def test_python_and_html_omitted(self):
        collected = do_collect()

        self.assertIn("safer_staticfiles/safer_staticfiles.css", collected["modified"])
        self.assertIn("safer_staticfiles/safer_staticfiles.js", collected["modified"])
        self.assertNotIn("safer_staticfiles/safer_staticfiles.html", collected["modified"])
        self.assertNotIn("safer_staticfiles/safer_staticfiles.py", collected["modified"])

        self.assertListEqual(collected["unmodified"], [])
        self.assertListEqual(collected["post_processed"], [])



from unittest.mock import Mock

from django.http import HttpResponseNotModified
from django.template import Template
from django.test import override_settings

# TODO - Convert to dc?
from django_components import component, types
from django_components.middleware import ComponentDependencyMiddleware

from .django_test_setup import setup_test_config
from .testutils import BaseTestCase, create_and_process_template_response

setup_test_config()


class SimpleComponent(component.Component):
    template: types.django_html = """
        Variable: <strong>{{ variable }}</strong>
    """

    def get_context_data(self, variable, variable2="default"):
        return {
            "variable": variable,
            "variable2": variable2,
        }

    class Media:
        css = "style.css"
        js = "script.js"


class SimpleComponentAlternate(component.Component):
    template: types.django_html = """
        Variable: <strong>{{ variable }}</strong>
    """

    def get_context_data(self, variable):
        return {}

    class Media:
        css = "style2.css"
        js = "script2.js"


class SimpleComponentWithSharedDependency(component.Component):
    template: types.django_html = """
        Variable: <strong>{{ variable }}</strong>
    """

    def get_context_data(self, variable, variable2="default"):
        return {}

    class Media:
        css = ["style.css", "style2.css"]
        js = ["script.js", "script2.js"]


class MultistyleComponent(component.Component):
    template: types.django_html = """
        Variable: <strong>{{ variable }}</strong>
    """

    class Media:
        css = ["style.css", "style2.css"]
        js = ["script.js", "script2.js"]


@override_settings(COMPONENTS={"RENDER_DEPENDENCIES": True})
class ComponentMediaRenderingTests(BaseTestCase):
    def setUp(self):
        # NOTE: component.registry is global, so need to clear before each test
        component.registry.clear()

    def test_no_dependencies_when_no_components_used(self):
        component.registry.register(name="test", component=SimpleComponent)

        template_str: types.django_html = """
            {% load component_tags %}{% component_dependencies %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML('<script src="script.js">', rendered, count=0)
        self.assertInHTML(
            '<link href="style.css" media="all" rel="stylesheet"/>',
            rendered,
            count=0,
        )

    def test_no_js_dependencies_when_no_components_used(self):
        component.registry.register(name="test", component=SimpleComponent)

        template_str: types.django_html = """
            {% load component_tags %}{% component_js_dependencies %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML('<script src="script.js">', rendered, count=0)

    def test_no_css_dependencies_when_no_components_used(self):
        component.registry.register(name="test", component=SimpleComponent)

        template_str: types.django_html = """
            {% load component_tags %}{% component_css_dependencies %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML(
            '<link href="style.css" media="all" rel="stylesheet"/>',
            rendered,
            count=0,
        )

    def test_preload_dependencies_render_when_no_components_used(self):
        component.registry.register(name="test", component=SimpleComponent)

        template_str: types.django_html = """
            {% load component_tags %}{% component_dependencies preload='test' %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML('<script src="script.js">', rendered, count=1)
        self.assertInHTML(
            '<link href="style.css" media="all" rel="stylesheet"/>',
            rendered,
            count=1,
        )

    def test_preload_css_dependencies_render_when_no_components_used(self):
        component.registry.register(name="test", component=SimpleComponent)

        template_str: types.django_html = """
            {% load component_tags %}{% component_css_dependencies preload='test' %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML(
            '<link href="style.css" media="all" rel="stylesheet"/>',
            rendered,
            count=1,
        )

    def test_single_component_dependencies_render_when_used(self):
        component.registry.register(name="test", component=SimpleComponent)

        template_str: types.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'test' variable='foo' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML(
            '<link href="style.css" media="all" rel="stylesheet"/>',
            rendered,
            count=1,
        )
        self.assertInHTML('<script src="script.js">', rendered, count=1)

    def test_single_component_with_dash_or_slash_in_name(self):
        component.registry.register(name="test", component=SimpleComponent)

        template_str: types.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'test' variable='foo' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML(
            '<link href="style.css" media="all" rel="stylesheet"/>',
            rendered,
            count=1,
        )
        self.assertInHTML('<script src="script.js">', rendered, count=1)

    def test_preload_dependencies_render_once_when_used(self):
        component.registry.register(name="test", component=SimpleComponent)

        template_str: types.django_html = """
            {% load component_tags %}{% component_dependencies preload='test' %}
            {% component 'test' variable='foo' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML(
            '<link href="style.css" media="all" rel="stylesheet"/>',
            rendered,
            count=1,
        )
        self.assertInHTML('<script src="script.js">', rendered, count=1)

    def test_placeholder_removed_when_single_component_rendered(self):
        component.registry.register(name="test", component=SimpleComponent)

        template_str: types.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'test' variable='foo' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertNotIn("_RENDERED", rendered)

    def test_placeholder_removed_when_preload_rendered(self):
        component.registry.register(name="test", component=SimpleComponent)

        template_str: types.django_html = """
            {% load component_tags %}{% component_dependencies preload='test' %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertNotIn("_RENDERED", rendered)

    def test_single_component_css_dependencies(self):
        component.registry.register(name="test", component=SimpleComponent)

        template_str: types.django_html = """
            {% load component_tags %}{% component_css_dependencies %}
            {% component 'test' variable='foo' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML(
            '<link href="style.css" media="all" rel="stylesheet"/>',
            rendered,
            count=1,
        )

    def test_single_component_js_dependencies(self):
        component.registry.register(name="test", component=SimpleComponent)

        template_str: types.django_html = """
            {% load component_tags %}{% component_js_dependencies %}
            {% component 'test' variable='foo' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML('<script src="script.js">', rendered, count=1)

    def test_all_dependencies_are_rendered_for_component_with_multiple_dependencies(
        self,
    ):
        component.registry.register(name="test", component=MultistyleComponent)
        template_str: types.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'test' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML('<script src="script.js">', rendered, count=1)
        self.assertInHTML('<script src="script2.js">', rendered, count=1)
        self.assertInHTML(
            '<link href="style.css" media="all" rel="stylesheet"/>',
            rendered,
            count=1,
        )
        self.assertInHTML(
            '<link href="style2.css" media="all" rel="stylesheet"/>',
            rendered,
            count=1,
        )

    def test_all_js_dependencies_are_rendered_for_component_with_multiple_dependencies(
        self,
    ):
        component.registry.register(name="test", component=MultistyleComponent)
        template_str: types.django_html = """
            {% load component_tags %}{% component_js_dependencies %}
            {% component 'test' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML('<script src="script.js">', rendered, count=1)
        self.assertInHTML('<script src="script2.js">', rendered, count=1)
        self.assertInHTML(
            '<link href="style.css" media="all" rel="stylesheet"/>',
            rendered,
            count=0,
        )
        self.assertInHTML(
            '<link href="style2.css" media="all" rel="stylesheet"/>',
            rendered,
            count=0,
        )

    def test_all_css_dependencies_are_rendered_for_component_with_multiple_dependencies(
        self,
    ):
        component.registry.register(name="test", component=MultistyleComponent)
        template_str: types.django_html = """
            {% load component_tags %}{% component_css_dependencies %}
            {% component 'test' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML('<script src="script.js">', rendered, count=0)
        self.assertInHTML('<script src="script2.js">', rendered, count=0)
        self.assertInHTML(
            '<link href="style.css" media="all" rel="stylesheet"/>',
            rendered,
            count=1,
        )
        self.assertInHTML(
            '<link href="style2.css" media="all" rel="stylesheet"/>',
            rendered,
            count=1,
        )

    def test_no_dependencies_with_multiple_unused_components(self):
        component.registry.register(name="test1", component=SimpleComponent)
        component.registry.register(name="test2", component=SimpleComponentAlternate)

        template_str: types.django_html = """
            {% load component_tags %}{% component_dependencies %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML('<script src="script.js">', rendered, count=0)
        self.assertInHTML('<script src="script2.js">', rendered, count=0)
        self.assertInHTML(
            '<link href="style.css" media="all" rel="stylesheet"/>',
            rendered,
            count=0,
        )
        self.assertInHTML(
            '<link href="style2.css" media="all" rel="stylesheet"/>',
            rendered,
            count=0,
        )

    def test_correct_css_dependencies_with_multiple_components(self):
        component.registry.register(name="test1", component=SimpleComponent)
        component.registry.register(name="test2", component=SimpleComponentAlternate)

        template_str: types.django_html = """
            {% load component_tags %}{% component_css_dependencies %}
            {% component 'test1' 'variable' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML(
            '<link href="style.css" media="all" rel="stylesheet"/>',
            rendered,
            count=1,
        )
        self.assertInHTML(
            '<link href="style2.css" media="all" rel="stylesheet"/>',
            rendered,
            count=0,
        )

    def test_correct_js_dependencies_with_multiple_components(self):
        component.registry.register(name="test1", component=SimpleComponent)
        component.registry.register(name="test2", component=SimpleComponentAlternate)

        template_str: types.django_html = """
            {% load component_tags %}{% component_js_dependencies %}
            {% component 'test1' 'variable' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML('<script src="script.js">', rendered, count=1)
        self.assertInHTML('<script src="script2.js">', rendered, count=0)

    def test_correct_dependencies_with_multiple_components(self):
        component.registry.register(name="test1", component=SimpleComponent)
        component.registry.register(name="test2", component=SimpleComponentAlternate)

        template_str: types.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'test2' variable='variable' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML('<script src="script.js">', rendered, count=0)
        self.assertInHTML('<script src="script2.js">', rendered, count=1)
        self.assertInHTML(
            '<link href="style.css" media="all" rel="stylesheet"/>',
            rendered,
            count=0,
        )
        self.assertInHTML(
            '<link href="style2.css" media="all" rel="stylesheet"/>',
            rendered,
            count=1,
        )

    def test_shared_dependencies_rendered_once(self):
        component.registry.register(name="test1", component=SimpleComponent)
        component.registry.register(name="test2", component=SimpleComponentAlternate)
        component.registry.register(name="test3", component=SimpleComponentWithSharedDependency)

        template_str: types.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'test1' variable='variable' %}{% endcomponent %}
            {% component 'test2' variable='variable' %}{% endcomponent %}
            {% component 'test1' variable='variable' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertInHTML('<script src="script.js">', rendered, count=1)
        self.assertInHTML('<script src="script2.js">', rendered, count=1)
        self.assertInHTML(
            '<link href="style.css" media="all" rel="stylesheet"/>',
            rendered,
            count=1,
        )
        self.assertInHTML(
            '<link href="style2.css" media="all" rel="stylesheet"/>',
            rendered,
            count=1,
        )

    def test_placeholder_removed_when_multiple_component_rendered(self):
        component.registry.register(name="test1", component=SimpleComponent)
        component.registry.register(name="test2", component=SimpleComponentAlternate)
        component.registry.register(name="test3", component=SimpleComponentWithSharedDependency)

        template_str: types.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'test1' variable='variable' %}{% endcomponent %}
            {% component 'test2' variable='variable' %}{% endcomponent %}
            {% component 'test1' variable='variable' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = create_and_process_template_response(template)
        self.assertNotIn("_RENDERED", rendered)

    def test_middleware_response_without_content_type(self):
        response = HttpResponseNotModified()
        middleware = ComponentDependencyMiddleware(get_response=lambda _: response)
        request = Mock()
        self.assertEqual(response, middleware(request=request))

    def test_middleware_response_with_components_with_slash_dash_and_underscore(
        self,
    ):
        component_names = [
            "test-component",
            "test/component",
            "test_component",
        ]
        for component_name in component_names:
            component.registry.register(name=component_name, component=SimpleComponent)
            template_str: types.django_html = f"""
                {{% load component_tags %}}
                {{% component_js_dependencies %}}
                {{% component_css_dependencies %}}
                {{% component '{component_name}' variable='value' %}}
                {{% endcomponent %}}
            """
            template = Template(template_str)
            rendered = create_and_process_template_response(template)
            self.assertHTMLEqual(
                rendered,
                (
                    '<script src="script.js"></script>'
                    '<link href="style.css" media="all" rel="stylesheet">'
                    "Variable: <strong>value</strong>\n"
                ),
            )

    # def test_dependencies_with_no_components(self):
    #     from django.template import Context
    #     template = Template(
    #         """
    #         {% load component_tags %}
    #         {% component_dependencies %}
    #         """
    #     )
    #     rendered = template.render(Context())
    #     print("RENDERD: ", rendered)
    #     # Assertions to ensure placeholders are properly replaced or removed
    #     self.assertNotIn(
    #         '<script name="JS_PLACEHOLDER"></script>', rendered, "JS placeholder should be removed"
    #     )
    #     self.assertNotIn(
    #         '<link name="CSS_PLACEHOLDER">', rendered, "CSS placeholder should be removed or replaced"
    #     )



from django.template import Context, Template, TemplateSyntaxError

import django_components as dc

from .django_test_setup import setup_test_config
from .testutils import BaseTestCase, parametrize_context_behavior

setup_test_config()


class ProvideTemplateTagTest(BaseTestCase):
    @parametrize_context_behavior(["django", "isolated"])
    def test_provide_basic(self):
        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("my_provide", "default")
                return {"var": var}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide "my_provide" key="hi" another=123 %}
                {% component "injectee" %}
                {% endcomponent %}
            {% endprovide %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <div> injected: DepInject(key='hi', another=123) </div>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_provide_access_keys_in_python(self):
        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> key: {{ key }} </div>
                <div> another: {{ another }} </div>
            """

            def get_context_data(self):
                my_provide = self.inject("my_provide")
                return {
                    "key": my_provide.key,
                    "another": my_provide.another,
                }

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide "my_provide" key="hi" another=123 %}
                {% component "injectee" %}
                {% endcomponent %}
            {% endprovide %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <div> key: hi </div>
            <div> another: 123 </div>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_provide_access_keys_in_django(self):
        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> key: {{ my_provide.key }} </div>
                <div> another: {{ my_provide.another }} </div>
            """

            def get_context_data(self):
                my_provide = self.inject("my_provide")
                return {
                    "my_provide": my_provide,
                }

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide "my_provide" key="hi" another=123 %}
                {% component "injectee" %}
                {% endcomponent %}
            {% endprovide %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <div> key: hi </div>
            <div> another: 123 </div>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_provide_does_not_leak(self):
        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("my_provide", "default")
                return {"var": var}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide "my_provide" key="hi" another=123 %}
            {% endprovide %}
            {% component "injectee" %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <div> injected: default </div>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_provide_empty(self):
        """Check provide tag with no kwargs"""

        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("my_provide", "default")
                return {"var": var}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide "my_provide" %}
                {% component "injectee" %}
                {% endcomponent %}
            {% endprovide %}
            {% component "injectee" %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <div> injected: DepInject() </div>
            <div> injected: default </div>
        """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_provide_no_inject(self):
        """Check that nothing breaks if we do NOT inject even if some data is provided"""

        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div></div>
            """

            def get_context_data(self):
                return {}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide "my_provide" key="hi" another=123 %}
                {% component "injectee" %}
                {% endcomponent %}
            {% endprovide %}
            {% component "injectee" %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <div></div>
            <div></div>
        """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_provide_key_single_quotes(self):
        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("my_provide", "default")
                return {"var": var}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide 'my_provide' key="hi" another=123 %}
                {% component "injectee" %}
                {% endcomponent %}
            {% endprovide %}
            {% component "injectee" %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <div> injected: DepInject(key='hi', another=123) </div>
            <div> injected: default </div>
        """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_provide_no_key_raises(self):
        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("my_provide", "default")
                return {"var": var}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide key="hi" another=123 %}
                {% component "injectee" %}
                {% endcomponent %}
            {% endprovide %}
            {% component "injectee" %}
            {% endcomponent %}
        """
        with self.assertRaises(TemplateSyntaxError):
            Template(template_str)

    @parametrize_context_behavior(["django", "isolated"])
    def test_provide_key_must_be_string_literal(self):
        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("my_provide", "default")
                return {"var": var}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide my_var key="hi" another=123 %}
                {% component "injectee" %}
                {% endcomponent %}
            {% endprovide %}
            {% component "injectee" %}
            {% endcomponent %}
        """
        with self.assertRaises(TemplateSyntaxError):
            Template(template_str)

    @parametrize_context_behavior(["django", "isolated"])
    def test_provide_key_must_be_identifier(self):
        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("my_provide", "default")
                return {"var": var}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide "%heya%" key="hi" another=123 %}
                {% component "injectee" %}
                {% endcomponent %}
            {% endprovide %}
            {% component "injectee" %}
            {% endcomponent %}
        """
        template = Template(template_str)
        with self.assertRaises(TemplateSyntaxError):
            template.render(Context({}))

    @parametrize_context_behavior(["django", "isolated"])
    def test_provide_aggregate_dics(self):
        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("my_provide", "default")
                return {"var": var}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide "my_provide" var1:key="hi" var1:another=123 var2:x="y" %}
                {% component "injectee" %}
                {% endcomponent %}
            {% endprovide %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <div> injected: DepInject(var1={'key': 'hi', 'another': 123}, var2={'x': 'y'}) </div>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_provide_does_not_expose_kwargs_to_context(self):
        """Check that `provide` tag doesn't assign the keys to the context like `with` tag does"""

        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("my_provide", "default")
                return {"var": var}

        template_str: dc.django_html = """
            {% load component_tags %}
            var_out: {{ var }}
            key_out: {{ key }}
            {% provide "my_provide" key="hi" another=123 %}
                var_in: {{ var }}
                key_in: {{ key }}
            {% endprovide %}
        """
        template = Template(template_str)
        rendered = template.render(Context({"var": "123"}))

        self.assertHTMLEqual(
            rendered,
            """
            var_out: 123
            key_out:
            var_in: 123
            key_in:
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_provide_nested_in_provide_same_key(self):
        """Check that inner `provide` with same key overshadows outer `provide`"""

        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("my_provide", "default")
                return {"var": var}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide "my_provide" key="hi" another=123 lost=0 %}
                {% provide "my_provide" key="hi1" another=1231 new=3 %}
                    {% component "injectee" %}
                    {% endcomponent %}
                {% endprovide %}

                {% component "injectee" %}
                {% endcomponent %}
            {% endprovide %}
            {% component "injectee" %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <div> injected: DepInject(key='hi1', another=1231, new=3) </div>
            <div> injected: DepInject(key='hi', another=123, lost=0) </div>
            <div> injected: default </div>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_provide_nested_in_provide_different_key(self):
        """Check that `provide` tag with different keys don't affect each other"""

        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> first_provide: {{ first_provide|safe }} </div>
                <div> second_provide: {{ second_provide|safe }} </div>
            """

            def get_context_data(self):
                first_provide = self.inject("first_provide", "default")
                second_provide = self.inject("second_provide", "default")
                return {
                    "first_provide": first_provide,
                    "second_provide": second_provide,
                }

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide "first_provide" key="hi" another=123 lost=0 %}
                {% provide "second_provide" key="hi1" another=1231 new=3 %}
                    {% component "injectee" %}
                    {% endcomponent %}
                {% endprovide %}
            {% endprovide %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <div> first_provide: DepInject(key='hi', another=123, lost=0) </div>
            <div> second_provide: DepInject(key='hi1', another=1231, new=3) </div>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_provide_in_include(self):
        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("my_provide", "default")
                return {"var": var}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide "my_provide" key="hi" another=123 %}
                {% include "inject.html" %}
            {% endprovide %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <div>
                <div> injected: DepInject(key='hi', another=123) </div>
            </div>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_slot_in_provide(self):
        @component.register("injectee")
        class InjectComponent(component.Component):
            template: types.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("my_provide", "default")
                return {"var": var}

        @component.register("parent")
        class ParentComponent(component.Component):
            template: types.django_html = """
                {% load component_tags %}
                {% provide "my_provide" key="hi" another=123 %}
                    {% slot "content" default %}{% endslot %}
                {% endprovide %}
            """

        template_str: types.django_html = """
            {% load component_tags %}
            {% component "parent" %}
                {% component "injectee" %}{% endcomponent %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <div>
                injected: DepInject(key='hi', another=123)
            </div>
            """,
        )


class InjectTest(BaseTestCase):
    @parametrize_context_behavior(["django", "isolated"])
    def test_inject_basic(self):
        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("my_provide")
                return {"var": var}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide "my_provide" key="hi" another=123 %}
                {% component "injectee" %}
                {% endcomponent %}
            {% endprovide %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <div> injected: DepInject(key='hi', another=123) </div>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_inject_missing_key_raises_without_default(self):
        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("abc")
                return {"var": var}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "injectee" %}
            {% endcomponent %}
        """
        template = Template(template_str)

        with self.assertRaises(KeyError):
            template.render(Context({}))

    @parametrize_context_behavior(["django", "isolated"])
    def test_inject_missing_key_ok_with_default(self):
        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("abc", "default")
                return {"var": var}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "injectee" %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))
        self.assertHTMLEqual(
            rendered,
            """
            <div> injected: default </div>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_inject_empty_string(self):
        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("")
                return {"var": var}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% provide "my_provide" key="hi" another=123 %}
                {% component "injectee" %}
                {% endcomponent %}
            {% endprovide %}
            {% component "injectee" %}
            {% endcomponent %}
        """
        template = Template(template_str)

        with self.assertRaises(KeyError):
            template.render(Context({}))

    @parametrize_context_behavior(["django", "isolated"])
    def test_inject_raises_on_called_outside_get_context_data(self):
        @dc.register("injectee")
        class InjectComponent(dc.Component):
            template: dc.django_html = """
                <div> injected: {{ var|safe }} </div>
            """

            def get_context_data(self):
                var = self.inject("abc", "default")
                return {"var": var}

        comp = InjectComponent("")
        with self.assertRaises(RuntimeError):
            comp.inject("abc", "def")



from django.template import Context, Template

import django_components as dc
from django_components.tag_formatter import ShorthandTagFormatter

from .django_test_setup import setup_test_config
from .testutils import BaseTestCase, parametrize_context_behavior


setup_test_config()


class MultiwordBlockStartTagFormatter(ShorthandTagFormatter):
    def format_block_start_tag(self, name):
        return f"{name} comp"


class MultiwordBlockEndTagFormatter(ShorthandTagFormatter):
    def format_block_end_tag(self, name):
        return f"end {name}"


class MultiwordInlineTagFormatter(ShorthandTagFormatter):
    def format_inline_tag(self, name):
        return f"#{name} comp"


# Create a TagFormatter class to validate the public interface
def create_validator_tag_formatter(tag_name: str):
    class ValidatorTagFormatter(ShorthandTagFormatter):
        def format_block_start_tag(self, name):
            assert name == tag_name
            return super().format_block_start_tag(name)

        def format_block_end_tag(self, name):
            assert name == tag_name
            return super().format_block_end_tag(name)

        def format_inline_tag(self, name):
            assert name == tag_name
            return super().format_inline_tag(name)

        def parse_block_start_tag(self, tokens):
            assert isinstance(tokens, list)
            assert tokens[0] == tag_name
            return super().parse_block_start_tag(tokens)

        def parse_inline_tag(self, tokens):
            assert isinstance(tokens, list)
            assert tokens[0] == f"#{tag_name}"
            return super().parse_inline_tag(tokens)

    return ValidatorTagFormatter


class ComponentTagTests(BaseTestCase):
    @parametrize_context_behavior(["django", "isolated"])
    def test_formatter_default_inline(self):
        @dc.register("simple")
        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                hello1
                <div>
                    {% slot "content" default %} SLOT_DEFAULT {% endslot %}
                </div>
                hello2
            """

        template = Template(
            """
            {% load component_tags %}
            {% #component "simple" %}
        """
        )
        rendered = template.render(Context())
        self.assertHTMLEqual(
            rendered,
            """
            hello1
            <div>
                SLOT_DEFAULT
            </div>
            hello2
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_formatter_default_block(self):
        @dc.register("simple")
        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                hello1
                <div>
                    {% slot "content" default %} SLOT_DEFAULT {% endslot %}
                </div>
                hello2
            """

        template = Template(
            """
            {% load component_tags %}
            {% component "simple" %}
                OVERRIDEN!
            {% endcomponent %}
        """
        )
        rendered = template.render(Context())
        self.assertHTMLEqual(
            rendered,
            """
            hello1
            <div>
                OVERRIDEN!
            </div>
            hello2
            """,
        )

    @parametrize_context_behavior(
        cases=["django", "isolated"],
        settings={
            "COMPONENTS": {
                "tag_formatter": "django_components.tag_formatter.ComponentTagFormatter",
            },
        },
    )
    def test_formatter_component_inline(self):
        @dc.register("simple")
        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                hello1
                <div>
                    {% slot "content" default %} SLOT_DEFAULT {% endslot %}
                </div>
                hello2
            """

        template = Template(
            """
            {% load component_tags %}
            {% #component "simple" %}
        """
        )
        rendered = template.render(Context())
        self.assertHTMLEqual(
            rendered,
            """
            hello1
            <div>
                SLOT_DEFAULT
            </div>
            hello2
            """,
        )

    @parametrize_context_behavior(
        cases=["django", "isolated"],
        settings={
            "COMPONENTS": {
                "tag_formatter": "django_components.tag_formatter.ComponentTagFormatter",
            },
        },
    )
    def test_formatter_component_block(self):
        @dc.register("simple")
        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                hello1
                <div>
                    {% slot "content" default %} SLOT_DEFAULT {% endslot %}
                </div>
                hello2
            """

        template = Template(
            """
            {% load component_tags %}
            {% component "simple" %}
                OVERRIDEN!
            {% endcomponent %}
        """
        )
        rendered = template.render(Context())
        self.assertHTMLEqual(
            rendered,
            """
            hello1
            <div>
                OVERRIDEN!
            </div>
            hello2
            """,
        )

    @parametrize_context_behavior(
        cases=["django", "isolated"],
        settings={
            "COMPONENTS": {
                "tag_formatter": "django_components.tag_formatter.ShorthandTagFormatter",
            },
        },
    )
    def test_formatter_shorthand_inline(self):
        @dc.register("simple")
        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                hello1
                <div>
                    {% slot "content" default %} SLOT_DEFAULT {% endslot %}
                </div>
                hello2
            """

        template = Template(
            """
            {% load component_tags %}
            {% #simple %}
        """
        )
        rendered = template.render(Context())
        self.assertHTMLEqual(
            rendered,
            """
            hello1
            <div>
                SLOT_DEFAULT
            </div>
            hello2
            """,
        )

    @parametrize_context_behavior(
        cases=["django", "isolated"],
        settings={
            "COMPONENTS": {
                "tag_formatter": "django_components.tag_formatter.ShorthandTagFormatter",
            },
        },
    )
    def test_formatter_shorthand_block(self):
        @dc.register("simple")
        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                hello1
                <div>
                    {% slot "content" default %} SLOT_DEFAULT {% endslot %}
                </div>
                hello2
            """

        template = Template(
            """
            {% load component_tags %}
            {% simple %}
                OVERRIDEN!
            {% endsimple %}
        """
        )
        rendered = template.render(Context())
        self.assertHTMLEqual(
            rendered,
            """
            hello1
            <div>
                OVERRIDEN!
            </div>
            hello2
            """,
        )

    @parametrize_context_behavior(
        cases=["django", "isolated"],
        settings={
            "COMPONENTS": {
                "tag_formatter": ShorthandTagFormatter,
            },
        },
    )
    def test_import_formatter_by_value(self):
        @dc.register("simple")
        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    {% slot "content" default %} SLOT_DEFAULT {% endslot %}
                </div>
            """

        template = Template(
            """
            {% load component_tags %}
            {% simple %}
                OVERRIDEN!
            {% endsimple %}
        """
        )
        rendered = template.render(Context())
        self.assertHTMLEqual(
            rendered,
            """
            <div>
                OVERRIDEN!
            </div>
            """,
        )

    @parametrize_context_behavior(
        cases=["django", "isolated"],
        settings={
            "COMPONENTS": {
                "tag_formatter": MultiwordBlockStartTagFormatter,
            },
        },
    )
    def test_raises_on_invalid_block_start_tag(self):
        with self.assertRaisesMessage(
            ValueError, "MultiwordBlockStartTagFormatter returned an invalid tag for block_start_tag: 'simple comp'"
        ):

            @dc.register("simple")
            class SimpleComponent(dc.Component):
                template = """{% load component_tags %}"""

    @parametrize_context_behavior(
        cases=["django", "isolated"],
        settings={
            "COMPONENTS": {
                "tag_formatter": MultiwordBlockEndTagFormatter,
            },
        },
    )
    def test_raises_on_invalid_block_end_tag(self):
        with self.assertRaisesMessage(
            ValueError, "MultiwordBlockEndTagFormatter returned an invalid tag for block_end_tag: 'end simple'"
        ):

            @dc.register("simple")
            class SimpleComponent(dc.Component):
                template: dc.django_html = """
                    {% load component_tags %}
                    <div>
                        {% slot "content" default %} SLOT_DEFAULT {% endslot %}
                    </div>
                """

            Template(
                """
                {% load component_tags %}
                {% simple %}
                    OVERRIDEN!
                {% bar %}
            """
            )

    @parametrize_context_behavior(
        cases=["django", "isolated"],
        settings={
            "COMPONENTS": {
                "tag_formatter": MultiwordInlineTagFormatter,
            },
        },
    )
    def test_raises_on_invalid_inline_tag(self):
        with self.assertRaisesMessage(
            ValueError, "MultiwordInlineTagFormatter returned an invalid tag for inline_tag: '#simple comp'"
        ):

            @dc.register("simple")
            class SimpleComponent(dc.Component):
                template: dc.django_html = """
                    {% load component_tags %}
                    <div>
                        {% slot "content" default %} SLOT_DEFAULT {% endslot %}
                    </div>
                """

    @parametrize_context_behavior(
        cases=["django", "isolated"],
        settings={
            "COMPONENTS": {
                "tag_formatter": create_validator_tag_formatter("simple"),
            },
        },
    )
    def test_method_args(self):
        @dc.register("simple")
        class SimpleComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                hello1
                <div>
                    {% slot "content" default %} SLOT_DEFAULT {% endslot %}
                </div>
                hello2
            """

        template = Template(
            """
            {% load component_tags %}
            {% #simple %}
        """
        )
        rendered = template.render(Context())
        self.assertHTMLEqual(
            rendered,
            """
            hello1
            <div>
                SLOT_DEFAULT
            </div>
            hello2
            """,
        )

        template = Template(
            """
            {% load component_tags %}
            {% simple %}
                OVERRIDEN!
            {% endsimple %}
        """
        )
        rendered = template.render(Context())
        self.assertHTMLEqual(
            rendered,
            """
            hello1
            <div>
                OVERRIDEN!
            </div>
            hello2
            """,
        )


# TODO:
# - Add tests for defining multiple components that produce the same tag, then render both
#   components in the same template. Then do `registry.clear`, try to render components now
#   (should fail), and then register the components again and render them again.
# - DOCUMENT CHANGED API
#   - types...
#   - component.Component -> dc.Component
#   - component.registry -> dc.registry
# - ADD TESTS TO COMPONENT REGISTRY
# - UPDATE README - API changes, tag_formatter, ComponentRegistry, Inline vs Block tags



from django.template import Context, Template, TemplateSyntaxError
from django.utils.safestring import SafeString, mark_safe

import django_components as dc
from django_components.attributes import append_attributes, attributes_to_string

from .django_test_setup import setup_test_config
from .testutils import BaseTestCase, parametrize_context_behavior

setup_test_config({"autodiscover": False})


class AttributesToStringTest(BaseTestCase):
    def test_simple_attribute(self):
        self.assertEqual(
            attributes_to_string({"foo": "bar"}),
            'foo="bar"',
        )

    def test_multiple_attributes(self):
        self.assertEqual(
            attributes_to_string({"class": "foo", "style": "color: red;"}),
            'class="foo" style="color: red;"',
        )

    def test_escapes_special_characters(self):
        self.assertEqual(
            attributes_to_string({"x-on:click": "bar", "@click": "'baz'"}),
            'x-on:click="bar" @click="&#x27;baz&#x27;"',
        )

    def test_does_not_escape_special_characters_if_safe_string(self):
        self.assertEqual(
            attributes_to_string({"foo": mark_safe("'bar'")}),
            "foo=\"'bar'\"",
        )

    def test_result_is_safe_string(self):
        result = attributes_to_string({"foo": mark_safe("'bar'")})
        self.assertTrue(isinstance(result, SafeString))

    def test_attribute_with_no_value(self):
        self.assertEqual(
            attributes_to_string({"required": None}),
            "",
        )

    def test_attribute_with_false_value(self):
        self.assertEqual(
            attributes_to_string({"required": False}),
            "",
        )

    def test_attribute_with_true_value(self):
        self.assertEqual(
            attributes_to_string({"required": True}),
            "required",
        )


class AppendAttributesTest(BaseTestCase):
    def test_single_dict(self):
        self.assertEqual(
            append_attributes(("foo", "bar")),
            {"foo": "bar"},
        )

    def test_appends_dicts(self):
        self.assertEqual(
            append_attributes(("class", "foo"), ("id", "bar"), ("class", "baz")),
            {"class": "foo baz", "id": "bar"},
        )


class HtmlAttrsTests(BaseTestCase):
    def setUp(self):
        super().setUp()

        self.template_str: dc.django_html = """
            {% load component_tags %}
            {% component "test" attrs:@click.stop="dispatch('click_event')" attrs:x-data="{hello: 'world'}" attrs:class=class_var %}
            {% endcomponent %}
        """  # noqa: E501

    @parametrize_context_behavior(["django", "isolated"])
    def test_tag_positional_args(self):
        @dc.register("test")
        class AttrsComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div {% html_attrs attrs defaults class="added_class" class="another-class" data-id=123 %}>
                    content
                </div>
            """  # noqa: E501

            def get_context_data(self, *args, attrs):
                return {
                    "attrs": attrs,
                    "defaults": {"class": "override-me"},
                }

        template = Template(self.template_str)
        rendered = template.render(Context({"class_var": "padding-top-8"}))
        self.assertHTMLEqual(
            rendered,
            """
            <div @click.stop="dispatch('click_event')" x-data="{hello: 'world'}" class="padding-top-8 added_class another-class" data-id=123>
                content
            </div>
            """,  # noqa: E501
        )
        self.assertNotIn("override-me", rendered)

    def test_tag_raises_on_extra_positional_args(self):
        @dc.register("test")
        class AttrsComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div {% html_attrs attrs defaults class %}>
                    content
                </div>
            """  # noqa: E501

            def get_context_data(self, *args, attrs):
                return {
                    "attrs": attrs,
                    "defaults": {"class": "override-me"},
                    "class": "123 457",
                }

        template = Template(self.template_str)

        with self.assertRaisesMessage(
            TemplateSyntaxError, "Tag 'html_attrs' received unexpected positional arguments"
        ):
            template.render(Context({"class_var": "padding-top-8"}))

    def test_tag_kwargs(self):
        @dc.register("test")
        class AttrsComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div {% html_attrs attrs=attrs defaults=defaults class="added_class" class="another-class" data-id=123 %}>
                    content
                </div>
            """  # noqa: E501

            def get_context_data(self, *args, attrs):
                return {
                    "attrs": attrs,
                    "defaults": {"class": "override-me"},
                }

        template = Template(self.template_str)
        rendered = template.render(Context({"class_var": "padding-top-8"}))
        self.assertHTMLEqual(
            rendered,
            """
            <div @click.stop="dispatch('click_event')" class="added_class another-class padding-top-8" data-id="123" x-data="{hello: 'world'}">
                content
            </div>
            """,  # noqa: E501
        )
        self.assertNotIn("override-me", rendered)

    def test_tag_kwargs_2(self):
        @dc.register("test")
        class AttrsComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div {% html_attrs class="added_class" class="another-class" data-id=123 defaults=defaults attrs=attrs %}>
                    content
                </div>
            """  # noqa: E501

            def get_context_data(self, *args, attrs):
                return {
                    "attrs": attrs,
                    "defaults": {"class": "override-me"},
                }

        template = Template(self.template_str)
        rendered = template.render(Context({"class_var": "padding-top-8"}))
        self.assertHTMLEqual(
            rendered,
            """
            <div @click.stop="dispatch('click_event')" x-data="{hello: 'world'}" class="padding-top-8 added_class another-class" data-id=123>
                content
            </div>
            """,  # noqa: E501
        )
        self.assertNotIn("override-me", rendered)

    def test_tag_aggregate_args(self):
        @dc.register("test")
        class AttrsComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div {% html_attrs attrs:class="from_agg_key" attrs:type="submit" defaults:class="override-me" class="added_class" class="another-class" data-id=123 %}>
                    content
                </div>
            """  # noqa: E501

            def get_context_data(self, *args, attrs):
                return {"attrs": attrs}

        template = Template(self.template_str)
        rendered = template.render(Context({"class_var": "padding-top-8"}))

        # NOTE: The attrs from self.template_str should be ignored because they are not used.
        self.assertHTMLEqual(
            rendered,
            """
            <div class="added_class another-class from_agg_key" data-id="123" type="submit">
                content
            </div>
            """,  # noqa: E501
        )
        self.assertNotIn("override-me", rendered)

    def test_tag_raises_on_aggregate_and_positional_args_for_attrs(self):
        @dc.register("test")
        class AttrsComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div {% html_attrs attrs attrs:class="from_agg_key" defaults:class="override-me" class="added_class" class="another-class" data-id=123 %}>
                    content
                </div>
            """  # noqa: E501

            def get_context_data(self, *args, attrs):
                return {"attrs": attrs}

        template = Template(self.template_str)

        with self.assertRaisesMessage(TemplateSyntaxError, "Received argument 'attrs' both as a regular input"):
            template.render(Context({"class_var": "padding-top-8"}))

    def test_tag_raises_on_aggregate_and_positional_args_for_defaults(self):
        @dc.register("test")
        class AttrsComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div {% html_attrs defaults=defaults attrs:class="from_agg_key" defaults:class="override-me" class="added_class" class="another-class" data-id=123 %}>
                    content
                </div>
            """  # noqa: E501

            def get_context_data(self, *args, attrs):
                return {"attrs": attrs}

        template = Template(self.template_str)

        with self.assertRaisesMessage(
            TemplateSyntaxError,
            "Received argument 'defaults' both as a regular input",
        ):
            template.render(Context({"class_var": "padding-top-8"}))

    def test_tag_no_attrs(self):
        @dc.register("test")
        class AttrsComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div {% html_attrs defaults:class="override-me" class="added_class" class="another-class" data-id=123 %}>
                    content
                </div>
            """  # noqa: E501

            def get_context_data(self, *args, attrs):
                return {"attrs": attrs}

        template = Template(self.template_str)
        rendered = template.render(Context({"class_var": "padding-top-8"}))
        self.assertHTMLEqual(
            rendered,
            """
            <div class="added_class another-class override-me" data-id=123>
                content
            </div>
            """,
        )

    def test_tag_no_defaults(self):
        @dc.register("test")
        class AttrsComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div {% html_attrs attrs class="added_class" class="another-class" data-id=123 %}>
                    content
                </div>
            """  # noqa: E501

            def get_context_data(self, *args, attrs):
                return {"attrs": attrs}

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "test" attrs:@click.stop="dispatch('click_event')" attrs:x-data="{hello: 'world'}" attrs:class=class_var %}
            {% endcomponent %}
        """  # noqa: E501
        template = Template(template_str)
        rendered = template.render(Context({"class_var": "padding-top-8"}))
        self.assertHTMLEqual(
            rendered,
            """
            <div @click.stop="dispatch('click_event')" x-data="{hello: 'world'}" class="padding-top-8 added_class another-class" data-id=123>
                content
            </div>
            """,  # noqa: E501
        )
        self.assertNotIn("override-me", rendered)

    def test_tag_no_attrs_no_defaults(self):
        @dc.register("test")
        class AttrsComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div {% html_attrs class="added_class" class="another-class" data-id=123 %}>
                    content
                </div>
            """  # noqa: E501

            def get_context_data(self, *args, attrs):
                return {"attrs": attrs}

        template = Template(self.template_str)
        rendered = template.render(Context({"class_var": "padding-top-8"}))
        self.assertHTMLEqual(
            rendered,
            """
            <div class="added_class another-class" data-id="123">
                content
            </div>
            """,
        )
        self.assertNotIn("override-me", rendered)

    def test_tag_empty(self):
        @dc.register("test")
        class AttrsComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div {% html_attrs %}>
                    content
                </div>
            """

            def get_context_data(self, *args, attrs):
                return {
                    "attrs": attrs,
                    "defaults": {"class": "override-me"},
                }

        template = Template(self.template_str)
        rendered = template.render(Context({"class_var": "padding-top-8"}))
        self.assertHTMLEqual(
            rendered,
            """
            <div >
                content
            </div>
            """,
        )
        self.assertNotIn("override-me", rendered)



from django.test import override_settings

from django_components.app_settings import app_settings

from .django_test_setup import setup_test_config
from .testutils import BaseTestCase

setup_test_config()


class SettingsTestCase(BaseTestCase):
    @override_settings(COMPONENTS={"context_behavior": "isolated"})
    def test_valid_context_behavior(self):
        self.assertEqual(app_settings.CONTEXT_BEHAVIOR, "isolated")

    @override_settings(COMPONENTS={"context_behavior": "invalid_value"})
    def test_raises_on_invalid_context_behavior(self):
        with self.assertRaises(ValueError):
            app_settings.CONTEXT_BEHAVIOR



from typing import Any, Dict, List, Optional

from django.template import Context, Template, TemplateSyntaxError

import django_components as dc

from .django_test_setup import setup_test_config
from .testutils import BaseTestCase, parametrize_context_behavior

setup_test_config()


class SlottedComponent(dc.Component):
    template: dc.django_html = """
        {% load component_tags %}
        <custom-template>
            <header>{% slot "header" %}Default header{% endslot %}</header>
            <main>{% slot "main" %}Default main{% endslot %}</main>
            <footer>{% slot "footer" %}Default footer{% endslot %}</footer>
        </custom-template>
    """


class SlottedComponentWithContext(SlottedComponent):
    def get_context_data(self, variable):
        return {"variable": variable}


#######################
# TESTS
#######################


class ComponentSlottedTemplateTagTest(BaseTestCase):
    def setUp(self):
        # NOTE: component.registry is global, so need to clear before each test
        dc.registry.clear()

    @parametrize_context_behavior(["django", "isolated"])
    def test_slotted_template_basic(self):
        dc.registry.register(name="test1", component=SlottedComponent)

        @dc.register("test2")
        class SimpleComponent(dc.Component):
            template = """Variable: <strong>{{ variable }}</strong>"""

            def get_context_data(self, variable, variable2="default"):
                return {
                    "variable": variable,
                    "variable2": variable2,
                }

            class Media:
                css = "style.css"
                js = "script.js"

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "test1" %}
                {% fill "header" %}
                    Custom header
                {% endfill %}
                {% fill "main" %}
                    {% component "test2" variable="variable" %}{% endcomponent %}
                {% endfill %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <custom-template>
                <header>Custom header</header>
                <main>Variable: <strong>variable</strong></main>
                <footer>Default footer</footer>
            </custom-template>
        """,
        )

    # NOTE: Second arg is the expected output of `{{ variable }}`
    @parametrize_context_behavior([("django", "test456"), ("isolated", "")])
    def test_slotted_template_with_context_var(self, context_behavior_data):
        dc.registry.register(name="test1", component=SlottedComponentWithContext)

        template_str: dc.django_html = """
            {% load component_tags %}
            {% with my_first_variable="test123" %}
                {% component "test1" variable="test456" %}
                    {% fill "main" %}
                        {{ my_first_variable }} - {{ variable }}
                    {% endfill %}
                    {% fill "footer" %}
                        {{ my_second_variable }}
                    {% endfill %}
                {% endcomponent %}
            {% endwith %}
        """
        template = Template(template_str)
        rendered = template.render(Context({"my_second_variable": "test321"}))

        self.assertHTMLEqual(
            rendered,
            f"""
            <custom-template>
                <header>Default header</header>
                <main>test123 - {context_behavior_data} </main>
                <footer>test321</footer>
            </custom-template>
        """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_slotted_template_no_slots_filled(self):
        dc.registry.register(name="test", component=SlottedComponent)

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "test" %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <custom-template>
                <header>Default header</header>
                <main>Default main</main>
                <footer>Default footer</footer>
            </custom-template>
        """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_slotted_template_without_slots(self):
        @dc.register("test")
        class SlottedComponentNoSlots(dc.Component):
            template: dc.django_html = """
                <custom-template></custom-template>
            """

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "test" %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(rendered, "<custom-template></custom-template>")

    @parametrize_context_behavior(["django", "isolated"])
    def test_slotted_template_without_slots_and_single_quotes(self):
        @dc.register("test")
        class SlottedComponentNoSlots(dc.Component):
            template: dc.django_html = """
                <custom-template></custom-template>
            """

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component 'test' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(rendered, "<custom-template></custom-template>")

    @parametrize_context_behavior(["django", "isolated"])
    def test_variable_fill_name(self):
        dc.registry.register(name="test", component=SlottedComponent)
        template_str: dc.django_html = """
            {% load component_tags %}
            {% with slotname="header" %}
                {% component 'test' %}
                    {% fill slotname %}Hi there!{% endfill %}
                {% endcomponent %}
            {% endwith %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))
        expected = """
        <custom-template>
            <header>Hi there!</header>
            <main>Default main</main>
            <footer>Default footer</footer>
        </custom-template>
        """
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_missing_required_slot_raises_error(self):
        class Component(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div class="header-box">
                    <h1>{% slot "title" required %}{% endslot %}</h1>
                    <h2>{% slot "subtitle" %}{% endslot %}</h2>
                </div>
            """

        dc.registry.register("test", Component)

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component 'test' %}
            {% endcomponent %}
        """
        template = Template(template_str)

        with self.assertRaises(TemplateSyntaxError):
            template.render(Context({}))

    @parametrize_context_behavior(["django", "isolated"])
    def test_default_slot_is_fillable_by_implicit_fill_content(self):
        @dc.register("test_comp")
        class ComponentWithDefaultSlot(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    <main>{% slot "main" default %}Easy to override{% endslot %}</main>
                </div>
            """

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component 'test_comp' %}
              <p>This fills the 'main' slot.</p>
            {% endcomponent %}
        """
        template = Template(template_str)

        expected = """
        <div>
          <main><p>This fills the 'main' slot.</p></main>
        </div>
        """
        rendered = template.render(Context({}))
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_default_slot_is_fillable_by_explicit_fill_content(self):
        @dc.register("test_comp")
        class ComponentWithDefaultSlot(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    <main>{% slot "main" default %}Easy to override{% endslot %}</main>
                </div>
            """

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component 'test_comp' %}
              {% fill "main" %}<p>This fills the 'main' slot.</p>{% endfill %}
            {% endcomponent %}
        """
        template = Template(template_str)
        expected = """
            <div>
            <main><p>This fills the 'main' slot.</p></main>
            </div>
        """
        rendered = template.render(Context({}))
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_error_raised_when_default_and_required_slot_not_filled(self):
        @dc.register("test_comp")
        class ComponentWithDefaultAndRequiredSlot(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    <header>{% slot "header" %}Your Header Here{% endslot %}</header>
                    <main>{% slot "main" default required %}Easy to override{% endslot %}</main>
                </div>
            """

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component 'test_comp' %}
            {% endcomponent %}
        """
        template = Template(template_str)

        with self.assertRaises(TemplateSyntaxError):
            template.render(Context({}))

    @parametrize_context_behavior(["django", "isolated"])
    def test_fill_tag_can_occur_within_component_nested_in_implicit_fill(self):
        dc.registry.register("slotted", SlottedComponent)

        @dc.register("test_comp")
        class ComponentWithDefaultSlot(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    <main>{% slot "main" default %}Easy to override{% endslot %}</main>
                </div>
            """

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component 'test_comp' %}
              {% component "slotted" %}
                {% fill "header" %}This Is Allowed{% endfill %}
                {% fill "main" %}{% endfill %}
                {% fill "footer" %}{% endfill %}
              {% endcomponent %}
            {% endcomponent %}
        """
        template = Template(template_str)
        expected = """
            <div>
            <main>
                <custom-template>
                <header>This Is Allowed</header>
                <main></main>
                <footer></footer>
                </custom-template>
            </main>
            </div>
        """
        rendered = template.render(Context({}))
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_error_from_mixed_implicit_and_explicit_fill_content(self):
        @dc.register("test_comp")
        class ComponentWithDefaultSlot(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    <main>{% slot "main" default %}Easy to override{% endslot %}</main>
                </div>
            """

        with self.assertRaises(TemplateSyntaxError):
            template_str: dc.django_html = """
                {% load component_tags %}
                {% component 'test_comp' %}
                  {% fill "main" %}Main content{% endfill %}
                  <p>And add this too!</p>
                {% endcomponent %}
            """
            Template(template_str)

    @parametrize_context_behavior(["django", "isolated"])
    def test_comments_permitted_inside_implicit_fill_content(self):
        @dc.register("test_comp")
        class ComponentWithDefaultSlot(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    <main>{% slot "main" default %}Easy to override{% endslot %}</main>
                </div>
            """

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component 'test_comp' %}
              <p>Main Content</p>
              {% comment %}
              This won't show up in the rendered HTML
              {% endcomment %}
              {# Nor will this #}
            {% endcomponent %}
        """
        Template(template_str)
        self.assertTrue(True)

    @parametrize_context_behavior(["django", "isolated"])
    def test_component_without_default_slot_refuses_implicit_fill(self):
        dc.registry.register("test_comp", SlottedComponent)
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component 'test_comp' %}
              <p>This shouldn't work because the included component doesn't mark
              any of its slots as 'default'</p>
            {% endcomponent %}
        """
        template = Template(template_str)

        with self.assertRaises(TemplateSyntaxError):
            template.render(Context({}))

    @parametrize_context_behavior(["django", "isolated"])
    def test_component_template_cannot_have_multiple_default_slots(self):
        class BadComponent(dc.Component):
            def get_template(self, context, template_name: Optional[str] = None) -> Template:
                template_str: dc.django_html = """
                    {% load django_components %}
                    <div>
                        {% slot "icon" %} {% endslot default %}
                        {% slot "description" %} {% endslot default %}
                    </div>
                """
                return Template(template_str)

        c = BadComponent("name")

        with self.assertRaises(TemplateSyntaxError):
            c.render(Context({}))

    @parametrize_context_behavior(["django", "isolated"])
    def test_slot_name_fill_typo_gives_helpful_error_message(self):
        dc.registry.register(name="test1", component=SlottedComponent)

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "test1" %}
                {% fill "haeder" %}
                    Custom header
                {% endfill %}
                {% fill "main" %}
                    main content
                {% endfill %}
            {% endcomponent %}
        """
        template = Template(template_str)

        with self.assertRaisesMessage(
            TemplateSyntaxError,
            (
                "Component 'test1' passed fill that refers to undefined slot: 'haeder'.\\n"
                "Unfilled slot names are: ['footer', 'header'].\\n"
                "Did you mean 'header'?"
            ),
        ):
            template.render(Context({}))

    # NOTE: This is relevant only for the "isolated" mode
    @parametrize_context_behavior(["isolated"])
    def test_slots_of_top_level_comps_can_access_full_outer_ctx(self):
        class SlottedComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    <main>{% slot "main" default %}Easy to override{% endslot %}</main>
                </div>
            """

            def get_context_data(self, name: Optional[str] = None) -> Dict[str, Any]:
                return {
                    "name": name,
                }

        dc.registry.register("test", SlottedComponent)

        template_str: dc.django_html = """
            {% load component_tags %}
            <body>
                {% component "test" %}
                    ABC: {{ name }} {{ some }}
                {% endcomponent %}
            </body>
        """
        self.template = Template(template_str)

        nested_ctx = Context()
        # Check that the component can access vars across different context layers
        nested_ctx.push({"some": "var"})
        nested_ctx.push({"name": "carl"})
        rendered = self.template.render(nested_ctx)

        self.assertHTMLEqual(
            rendered,
            """
            <body>
                <div>
                    <main> ABC: carl var </main>
                </div>
            </body>
            """,
        )


class SlottedTemplateRegressionTests(BaseTestCase):
    def setUp(self):
        # NOTE: component.registry is global, so need to clear before each test
        dc.registry.clear()

    @parametrize_context_behavior(["django", "isolated"])
    def test_slotted_template_that_uses_missing_variable(self):
        @dc.register("test")
        class SlottedComponentWithMissingVariable(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <custom-template>
                    {{ missing_context_variable }}
                    <header>{% slot "header" %}Default header{% endslot %}</header>
                    <main>{% slot "main" %}Default main{% endslot %}</main>
                    <footer>{% slot "footer" %}Default footer{% endslot %}</footer>
                </custom-template>
            """

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component 'test' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <custom-template>
                <header>Default header</header>
                <main>Default main</main>
                <footer>Default footer</footer>
            </custom-template>
            """,
        )


class SlotDefaultTests(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        dc.registry.clear()
        dc.registry.register("test", SlottedComponent)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        dc.registry.clear()

    @parametrize_context_behavior(["django", "isolated"])
    def test_basic(self):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "test" %}
                {% fill "header" default="header" %}Before: {{ header }}{% endfill %}
                {% fill "main" default="main" %}{{ main }}{% endfill %}
                {% fill "footer" default="footer" %}{{ footer }}, after{% endfill %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <custom-template>
                <header>Before: Default header</header>
                <main>Default main</main>
                <footer>Default footer, after</footer>
            </custom-template>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_multiple_calls(self):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "test" %}
                {% fill "header" default="header" %}
                    First: {{ header }};
                    Second: {{ header }}
                {% endfill %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <custom-template>
                <header>First: Default header; Second: Default header</header>
                <main>Default main</main>
                <footer>Default footer</footer>
            </custom-template>
        """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_under_if_and_forloop(self):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "test" %}
                {% fill "header" default="header" %}
                    {% for i in range %}
                        {% if forloop.first %}
                            First {{ header }}
                        {% else %}
                            Later {{ header }}
                        {% endif %}
                    {% endfor %}
                {% endfill %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({"range": range(3)}))

        self.assertHTMLEqual(
            rendered,
            """
            <custom-template>
                <header>First Default header Later Default header Later Default header</header>
                <main>Default main</main>
                <footer>Default footer</footer>
            </custom-template>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_nested_fills(self):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "test" %}
                {% fill "header" default="header1" %}
                    header1_in_header1: {{ header1 }}
                    {% component "test" %}
                        {% fill "header" default="header2" %}
                            header1_in_header2: {{ header1 }}
                            header2_in_header2: {{ header2 }}
                        {% endfill %}
                        {% fill "footer" default="footer2" %}
                            header1_in_footer2: {{ header1 }}
                            footer2_in_footer2: {{ footer2 }}
                        {% endfill %}
                    {% endcomponent %}
                {% endfill %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <custom-template>
                <header>
                    header1_in_header1: Default header
                    <custom-template>
                        <header>
                            header1_in_header2: Default header
                            header2_in_header2: Default header
                        </header>
                        <main>Default main</main>
                        <footer>
                            header1_in_footer2: Default header
                            footer2_in_footer2: Default footer
                        </footer>
                    </custom-template>
                </header>
                <main>Default main</main>
                <footer>Default footer</footer>
            </custom-template>
            """,
        )


class ScopedSlotTest(BaseTestCase):
    @parametrize_context_behavior(["django", "isolated"])
    def test_slot_data(self):
        @dc.register("test")
        class TestComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    {% slot "my_slot" abc=abc var123=var123 %}Default text{% endslot %}
                </div>
            """

            def get_context_data(self):
                return {
                    "abc": "def",
                    "var123": 456,
                }

        template: dc.django_html = """
            {% load component_tags %}
            {% component "test" %}
                {% fill "my_slot" data="slot_data_in_fill" %}
                    {{ slot_data_in_fill.abc }}
                    {{ slot_data_in_fill.var123 }}
                {% endfill %}
            {% endcomponent %}
        """
        rendered = Template(template).render(Context())
        expected = """
            <div>
                def
                456
            </div>
        """
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_slot_data_with_flags(self):
        @dc.register("test")
        class TestComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    {% slot "my_slot" default abc=abc var123=var123 required %}Default text{% endslot %}
                </div>
            """

            def get_context_data(self):
                return {
                    "abc": "def",
                    "var123": 456,
                }

        template: dc.django_html = """
            {% load component_tags %}
            {% component "test" %}
                {% fill "my_slot" data="slot_data_in_fill" %}
                    {{ slot_data_in_fill.abc }}
                    {{ slot_data_in_fill.var123 }}
                {% endfill %}
            {% endcomponent %}
        """
        rendered = Template(template).render(Context())
        expected = """
            <div>
                def
                456
            </div>
        """
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_slot_data_with_slot_default(self):
        @dc.register("test")
        class TestComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    {% slot "my_slot" abc=abc var123=var123 %}Default text{% endslot %}
                </div>
            """

            def get_context_data(self):
                return {
                    "abc": "def",
                    "var123": 456,
                }

        template: dc.django_html = """
            {% load component_tags %}
            {% component "test" %}
                {% fill "my_slot" data="slot_data_in_fill" default="slot_var" %}
                    {{ slot_var }}
                    {{ slot_data_in_fill.abc }}
                    {{ slot_data_in_fill.var123 }}
                {% endfill %}
            {% endcomponent %}
        """
        rendered = Template(template).render(Context())
        expected = """
            <div>
                Default text
                def
                456
            </div>
        """
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_slot_data_raises_on_slot_data_and_slot_default_same_var(self):
        @dc.register("test")
        class TestComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    {% slot "my_slot" abc=abc var123=var123 %}Default text{% endslot %}
                </div>
            """

            def get_context_data(self):
                return {
                    "abc": "def",
                    "var123": 456,
                }

        template: dc.django_html = """
            {% load component_tags %}
            {% component "test" %}
                {% fill "my_slot" data="slot_var" default="slot_var" %}
                    {{ slot_var }}
                {% endfill %}
            {% endcomponent %}
        """
        with self.assertRaisesMessage(
            TemplateSyntaxError,
            "'fill' received the same string for slot default (default=...) and slot data (data=...)",
        ):
            Template(template).render(Context())

    @parametrize_context_behavior(["django", "isolated"])
    def test_slot_data_fill_without_data(self):
        @dc.register("test")
        class TestComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    {% slot "my_slot" abc=abc var123=var123 %}Default text{% endslot %}
                </div>
            """

            def get_context_data(self):
                return {
                    "abc": "def",
                    "var123": 456,
                }

        template: dc.django_html = """
            {% load component_tags %}
            {% component "test" %}
                {% fill "my_slot" %}
                    overriden
                {% endfill %}
            {% endcomponent %}
        """
        rendered = Template(template).render(Context())
        expected = "<div> overriden </div>"
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_slot_data_fill_without_slot_data(self):
        @dc.register("test")
        class TestComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    {% slot "my_slot" %}Default text{% endslot %}
                </div>
            """

        template: dc.django_html = """
            {% load component_tags %}
            {% component "test" %}
                {% fill "my_slot" data="data" %}
                    {{ data|safe }}
                {% endfill %}
            {% endcomponent %}
        """
        rendered = Template(template).render(Context())
        expected = "<div> {} </div>"
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_slot_data_no_fill(self):
        @dc.register("test")
        class TestComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    {% slot "my_slot" abc=abc var123=var123 %}Default text{% endslot %}
                </div>
            """

            def get_context_data(self):
                return {
                    "abc": "def",
                    "var123": 456,
                }

        template: dc.django_html = """
            {% load component_tags %}
            {% component "test" %}
            {% endcomponent %}
        """
        rendered = Template(template).render(Context())
        expected = "<div> Default text </div>"
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_nested_fills(self):
        @dc.register("test")
        class TestComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <div>
                    {% slot "my_slot" abc=abc input=input %}Default text{% endslot %}
                </div>
            """

            def get_context_data(self, input):
                return {
                    "abc": "def",
                    "input": input,
                }

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "test" input=1 %}
                {% fill "my_slot" data="data1" %}
                    data1_in_slot1: {{ data1|safe }}
                    {% component "test" input=2 %}
                        {% fill "my_slot" data="data2" %}
                            data1_in_slot2: {{ data1|safe }}
                            data2_in_slot2: {{ data2|safe }}
                        {% endfill %}
                    {% endcomponent %}
                {% endfill %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({}))

        self.assertHTMLEqual(
            rendered,
            """
            <div>
                data1_in_slot1: {'abc': 'def', 'input': 1}
                <div>
                    data1_in_slot2: {'abc': 'def', 'input': 1}
                    data2_in_slot2: {'abc': 'def', 'input': 2}
                </div>
            </div>
            """,
        )


class DuplicateSlotTest(BaseTestCase):
    class DuplicateSlotComponent(dc.Component):
        template: dc.django_html = """
            {% load component_tags %}
            <header>{% slot "header" %}Default header{% endslot %}</header>
             {# Slot name 'header' used twice. #}
            <main>{% slot "header" %}Default main header{% endslot %}</main>
            <footer>{% slot "footer" %}Default footer{% endslot %}</footer>
        """

        def get_context_data(self, name: Optional[str] = None) -> Dict[str, Any]:
            return {
                "name": name,
            }

    class DuplicateSlotNestedComponent(dc.Component):
        template: dc.django_html = """
            {% load component_tags %}
            {% slot "header" %}START{% endslot %}
            <div class="dashboard-component">
            {% component "calendar" date="2020-06-06" %}
                {% fill "header" %}  {# fills and slots with same name relate to diff. things. #}
                    {% slot "header" %}NESTED{% endslot %}
                {% endfill %}
                {% fill "body" %}Here are your to-do items for today:{% endfill %}
            {% endcomponent %}
            <ol>
                {% for item in items %}
                    <li>{{ item }}</li>
                    {% slot "header" %}LOOP {{ item }} {% endslot %}
                {% endfor %}
            </ol>
            </div>
        """

        def get_context_data(self, items: List) -> Dict[str, Any]:
            return {
                "items": items,
            }

    class CalendarComponent(dc.Component):
        """Nested in ComponentWithNestedComponent"""

        template: dc.django_html = """
            {% load component_tags %}
            <div class="calendar-component">
            <h1>
                {% slot "header" %}Today's date is <span>{{ date }}</span>{% endslot %}
            </h1>
            <main>
                {% slot "body" %}
                    You have no events today.
                {% endslot %}
            </main>
            </div>
        """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        dc.registry.register(name="duplicate_slot", component=cls.DuplicateSlotComponent)
        dc.registry.register(name="duplicate_slot_nested", component=cls.DuplicateSlotNestedComponent)
        dc.registry.register(name="calendar", component=cls.CalendarComponent)

    # NOTE: Second arg is the input for the "name" component kwarg
    @parametrize_context_behavior(
        [
            # In "django" mode, we MUST pass name as arg through the component
            ("django", "Jannete"),
            # In "isolated" mode, the fill is already using top-level's context, so we pass nothing
            ("isolated", None),
        ]
    )
    def test_duplicate_slots(self, context_behavior_data):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "duplicate_slot" name=comp_input %}
                {% fill "header" %}
                    Name: {{ name }}
                {% endfill %}
                {% fill "footer" %}
                    Hello
                {% endfill %}
            {% endcomponent %}
        """
        self.template = Template(template_str)

        rendered = self.template.render(Context({"name": "Jannete", "comp_input": context_behavior_data}))
        self.assertHTMLEqual(
            rendered,
            """
            <header>Name: Jannete</header>
            <main>Name: Jannete</main>
            <footer>Hello</footer>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_duplicate_slots_fallback(self):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "duplicate_slot" %}
            {% endcomponent %}
        """
        self.template = Template(template_str)
        rendered = self.template.render(Context({}))

        # NOTE: Slots should have different fallbacks even though they use the same name
        self.assertHTMLEqual(
            rendered,
            """
            <header>Default header</header>
            <main>Default main header</main>
            <footer>Default footer</footer>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_duplicate_slots_nested(self):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "duplicate_slot_nested" items=items %}
                {% fill "header" %}
                    OVERRIDDEN!
                {% endfill %}
            {% endcomponent %}
        """
        self.template = Template(template_str)
        rendered = self.template.render(Context({"items": [1, 2, 3]}))

        # NOTE: Slots should have different fallbacks even though they use the same name
        self.assertHTMLEqual(
            rendered,
            """
            OVERRIDDEN!
            <div class="dashboard-component">
                <div class="calendar-component">
                    <h1>
                        OVERRIDDEN!
                    </h1>
                    <main>
                        Here are your to-do items for today:
                    </main>
                </div>

                <ol>
                    <li>1</li>
                    OVERRIDDEN!
                    <li>2</li>
                    OVERRIDDEN!
                    <li>3</li>
                    OVERRIDDEN!
                </ol>
            </div>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_duplicate_slots_nested_fallback(self):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "duplicate_slot_nested" items=items %}
            {% endcomponent %}
        """
        self.template = Template(template_str)
        rendered = self.template.render(Context({"items": [1, 2, 3]}))

        # NOTE: Slots should have different fallbacks even though they use the same name
        self.assertHTMLEqual(
            rendered,
            """
            START
            <div class="dashboard-component">
                <div class="calendar-component">
                    <h1>
                        NESTED
                    </h1>
                    <main>
                        Here are your to-do items for today:
                    </main>
                </div>

                <ol>
                    <li>1</li>
                    LOOP 1
                    <li>2</li>
                    LOOP 2
                    <li>3</li>
                    LOOP 3
                </ol>
            </div>
            """,
        )


class SlotFillTemplateSyntaxErrorTests(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        dc.registry.register("test", SlottedComponent)

    @classmethod
    def tearDownClass(cls) -> None:
        super().tearDownClass()
        dc.registry.clear()

    @parametrize_context_behavior(["django", "isolated"])
    def test_fill_with_no_parent_is_error(self):
        with self.assertRaises(TemplateSyntaxError):
            template_str: dc.django_html = """
                {% load component_tags %}
                {% fill "header" %}contents{% endfill %}
            """
            Template(template_str).render(Context({}))

    @parametrize_context_behavior(["django", "isolated"])
    def test_isolated_slot_is_error(self):
        @dc.register("broken_component")
        class BrokenComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                {% include 'slotted_template.html' with context=None only %}
            """

        template_str: types.django_html = """
            {% load component_tags %}
            {% component "broken_component" %}
                {% fill "header" %}Custom header {% endfill %}
                {% fill "main" %}Custom main{% endfill %}
                {% fill "footer" %}Custom footer{% endfill %}
            {% endcomponent %}
        """

        with self.assertRaises(KeyError):
            Template(template_str).render(Context({}))

    @parametrize_context_behavior(["django", "isolated"])
    def test_non_unique_fill_names_is_error(self):
        with self.assertRaises(TemplateSyntaxError):
            template_str: dc.django_html = """
                {% load component_tags %}
                {% component "test" %}
                    {% fill "header" %}Custom header {% endfill %}
                    {% fill "header" %}Other header{% endfill %}
                {% endcomponent %}
            """
            Template(template_str).render(Context({}))

    @parametrize_context_behavior(["django", "isolated"])
    def test_non_unique_fill_names_is_error_via_vars(self):
        with self.assertRaises(TemplateSyntaxError):
            template_str: dc.django_html = """
                {% load component_tags %}
                {% with var1="header" var2="header" %}
                    {% component "test" %}
                        {% fill var1 %}Custom header {% endfill %}
                        {% fill var2 %}Other header{% endfill %}
                    {% endcomponent %}
                {% endwith %}
            """
            Template(template_str).render(Context({}))


class SlotBehaviorTests(BaseTestCase):
    # NOTE: This is standalone function instead of setUp, so we can configure
    # Django settings per test with `@override_settings`
    def make_template(self) -> Template:
        class SlottedComponent(dc.Component):
            template: dc.django_html = """
                {% load component_tags %}
                <custom-template>
                    <header>{% slot "header" %}Default header{% endslot %}</header>
                    <main>{% slot "main" %}Default main{% endslot %}</main>
                    <footer>{% slot "footer" %}Default footer{% endslot %}</footer>
                </custom-template>
            """

            def get_context_data(self, name: Optional[str] = None) -> Dict[str, Any]:
                return {
                    "name": name,
                }

        dc.registry.register("test", SlottedComponent)

        template_str: dc.django_html = """
            {% load component_tags %}
            {% component "test" name='Igor' %}
                {% fill "header" %}
                    Name: {{ name }}
                {% endfill %}
                {% fill "main" %}
                    Day: {{ day }}
                {% endfill %}
                {% fill "footer" %}
                    {% component "test" name='Joe2' %}
                        {% fill "header" %}
                            Name2: {{ name }}
                        {% endfill %}
                        {% fill "main" %}
                            Day2: {{ day }}
                        {% endfill %}
                    {% endcomponent %}
                {% endfill %}
            {% endcomponent %}
        """
        return Template(template_str)

    @parametrize_context_behavior(["django"])
    def test_slot_context__django(self):
        template = self.make_template()
        # {{ name }} should be neither Jannete not empty, because overriden everywhere
        rendered = template.render(Context({"day": "Monday", "name": "Jannete"}))
        self.assertHTMLEqual(
            rendered,
            """
            <custom-template>
                <header>Name: Igor</header>
                <main>Day: Monday</main>
                <footer>
                    <custom-template>
                        <header>Name2: Joe2</header>
                        <main>Day2: Monday</main>
                        <footer>Default footer</footer>
                    </custom-template>
                </footer>
            </custom-template>
            """,
        )

        # {{ name }} should be effectively the same as before, because overriden everywhere
        rendered2 = template.render(Context({"day": "Monday"}))
        self.assertHTMLEqual(rendered2, rendered)

    @parametrize_context_behavior(["isolated"])
    def test_slot_context__isolated(self):
        template = self.make_template()
        # {{ name }} should be "Jannete" everywhere
        rendered = template.render(Context({"day": "Monday", "name": "Jannete"}))
        self.assertHTMLEqual(
            rendered,
            """
            <custom-template>
                <header>Name: Jannete</header>
                <main>Day: Monday</main>
                <footer>
                    <custom-template>
                        <header>Name2: Jannete</header>
                        <main>Day2: Monday</main>
                        <footer>Default footer</footer>
                    </custom-template>
                </footer>
            </custom-template>
            """,
        )

        # {{ name }} should be empty everywhere
        rendered2 = template.render(Context({"day": "Monday"}))
        self.assertHTMLEqual(
            rendered2,
            """
            <custom-template>
                <header>Name: </header>
                <main>Day: Monday</main>
                <footer>
                    <custom-template>
                        <header>Name2: </header>
                        <main>Day2: Monday</main>
                        <footer>Default footer</footer>
                    </custom-template>
                </footer>
            </custom-template>
            """,
        )



from django.template import Context, Template

import django_components as dc

from .django_test_setup import setup_test_config
from .testutils import BaseTestCase, parametrize_context_behavior

setup_test_config()


#########################
# COMPONENTS
#########################

class SimpleComponent(dc.Component):
    template: dc.django_html = """
        Variable: <strong>{{ variable }}</strong>
    """

    def get_context_data(self, variable=None):
        return {"variable": variable} if variable is not None else {}

    @staticmethod
    def expected_output(variable_value):
        return "Variable: < strong > {} < / strong >".format(variable_value)


class VariableDisplay(dc.Component):
    template: dc.django_html = """
        {% load component_tags %}
        <h1>Shadowing variable = {{ shadowing_variable }}</h1>
        <h1>Uniquely named variable = {{ unique_variable }}</h1>
    """

    def get_context_data(self, shadowing_variable=None, new_variable=None):
        context = {}
        if shadowing_variable is not None:
            context["shadowing_variable"] = shadowing_variable
        if new_variable is not None:
            context["unique_variable"] = new_variable
        return context


class IncrementerComponent(dc.Component):
    template: dc.django_html = """
        {% load component_tags %}
        <p class="incrementer">value={{ value }};calls={{ calls }}</p>
        {% slot 'content' %}{% endslot %}
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.call_count = 0

    def get_context_data(self, value=0):
        value = int(value)
        if hasattr(self, "call_count"):
            self.call_count += 1
        else:
            self.call_count = 1
        return {"value": value + 1, "calls": self.call_count}


#########################
# TESTS
#########################


class ContextTests(BaseTestCase):
    class ParentComponent(dc.Component):
        template: dc.django_html = """
            {% load component_tags %}
            <div>
                <h1>Parent content</h1>
                {% component name="variable_display" shadowing_variable='override' new_variable='unique_val' %}
                {% endcomponent %}
            </div>
            <div>
                {% slot 'content' %}
                    <h2>Slot content</h2>
                    {% component name="variable_display" shadowing_variable='slot_default_override' new_variable='slot_default_unique' %}
                    {% endcomponent %}
                {% endslot %}
            </div>
        """  # noqa

        def get_context_data(self):
            return {"shadowing_variable": "NOT SHADOWED"}

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        dc.registry.register(name="variable_display", component=VariableDisplay)
        dc.registry.register(name="parent_component", component=cls.ParentComponent)

    @parametrize_context_behavior(["django", "isolated"])
    def test_nested_component_context_shadows_parent_with_unfilled_slots_and_component_tag(
        self,
    ):
        template_str: dc.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'parent_component' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context())

        self.assertIn("<h1>Shadowing variable = override</h1>", rendered, rendered)
        self.assertIn(
            "<h1>Shadowing variable = slot_default_override</h1>",
            rendered,
            rendered,
        )
        self.assertNotIn("<h1>Shadowing variable = NOT SHADOWED</h1>", rendered, rendered)

    @parametrize_context_behavior(["django", "isolated"])
    def test_nested_component_instances_have_unique_context_with_unfilled_slots_and_component_tag(
        self,
    ):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component_dependencies %}
            {% component name='parent_component' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context())

        self.assertIn("<h1>Uniquely named variable = unique_val</h1>", rendered, rendered)
        self.assertIn(
            "<h1>Uniquely named variable = slot_default_unique</h1>",
            rendered,
            rendered,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_nested_component_context_shadows_parent_with_filled_slots(self):
        template_str: dc.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'parent_component' %}
                {% fill 'content' %}
                    {% component name='variable_display' shadowing_variable='shadow_from_slot' new_variable='unique_from_slot' %}
                    {% endcomponent %}
                {% endfill %}
            {% endcomponent %}
        """  # NOQA
        template = Template(template_str)
        rendered = template.render(Context())

        self.assertIn("<h1>Shadowing variable = override</h1>", rendered, rendered)
        self.assertIn(
            "<h1>Shadowing variable = shadow_from_slot</h1>",
            rendered,
            rendered,
        )
        self.assertNotIn("<h1>Shadowing variable = NOT SHADOWED</h1>", rendered, rendered)

    @parametrize_context_behavior(["django", "isolated"])
    def test_nested_component_instances_have_unique_context_with_filled_slots(self):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component_dependencies %}
            {% component 'parent_component' %}
                {% fill 'content' %}
                    {% component name='variable_display' shadowing_variable='shadow_from_slot' new_variable='unique_from_slot' %}
                    {% endcomponent %}
                {% endfill %}
            {% endcomponent %}
        """  # NOQA
        template = Template(template_str)
        rendered = template.render(Context())

        self.assertIn("<h1>Uniquely named variable = unique_val</h1>", rendered, rendered)
        self.assertIn(
            "<h1>Uniquely named variable = unique_from_slot</h1>",
            rendered,
            rendered,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_nested_component_context_shadows_outer_context_with_unfilled_slots_and_component_tag(
        self,
    ):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component_dependencies %}
            {% component name='parent_component' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({"shadowing_variable": "NOT SHADOWED"}))

        self.assertIn("<h1>Shadowing variable = override</h1>", rendered, rendered)
        self.assertIn(
            "<h1>Shadowing variable = slot_default_override</h1>",
            rendered,
            rendered,
        )
        self.assertNotIn("<h1>Shadowing variable = NOT SHADOWED</h1>", rendered, rendered)

    @parametrize_context_behavior(["django", "isolated"])
    def test_nested_component_context_shadows_outer_context_with_filled_slots(
        self,
    ):
        template_str: dc.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'parent_component' %}
                {% fill 'content' %}
                    {% component name='variable_display' shadowing_variable='shadow_from_slot' new_variable='unique_from_slot' %}
                    {% endcomponent %}
                {% endfill %}
            {% endcomponent %}
        """  # NOQA
        template = Template(template_str)
        rendered = template.render(Context({"shadowing_variable": "NOT SHADOWED"}))

        self.assertIn("<h1>Shadowing variable = override</h1>", rendered, rendered)
        self.assertIn(
            "<h1>Shadowing variable = shadow_from_slot</h1>",
            rendered,
            rendered,
        )
        self.assertNotIn("<h1>Shadowing variable = NOT SHADOWED</h1>", rendered, rendered)


class ParentArgsTests(BaseTestCase):
    class ParentComponentWithArgs(dc.Component):
        template: dc.django_html = """
            {% load component_tags %}
            <div>
                <h1>Parent content</h1>
                {% component name="variable_display" shadowing_variable=inner_parent_value new_variable='unique_val' %}
                {% endcomponent %}
            </div>
            <div>
                {% slot 'content' %}
                    <h2>Slot content</h2>
                    {% component name="variable_display" shadowing_variable='slot_default_override' new_variable=inner_parent_value %}
                    {% endcomponent %}
                {% endslot %}
            </div>
        """  # noqa

        def get_context_data(self, parent_value):
            return {"inner_parent_value": parent_value}

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        dc.registry.register(name="incrementer", component=IncrementerComponent)
        dc.registry.register(name="parent_with_args", component=cls.ParentComponentWithArgs)
        dc.registry.register(name="variable_display", component=VariableDisplay)

    @parametrize_context_behavior(["django", "isolated"])
    def test_parent_args_can_be_drawn_from_context(self):
        template_str: dc.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'parent_with_args' parent_value=parent_value %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({"parent_value": "passed_in"}))

        self.assertHTMLEqual(
            rendered,
            """
            <div>
                <h1>Parent content</h1>
                <h1>Shadowing variable = passed_in</h1>
                <h1>Uniquely named variable = unique_val</h1>
            </div>
            <div>
                <h2>Slot content</h2>
                <h1>Shadowing variable = slot_default_override</h1>
                <h1>Uniquely named variable = passed_in</h1>
            </div>
            """,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_parent_args_available_outside_slots(self):
        template_str: dc.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'parent_with_args' parent_value='passed_in' %}{%endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context())

        self.assertIn("<h1>Shadowing variable = passed_in</h1>", rendered, rendered)
        self.assertIn("<h1>Uniquely named variable = passed_in</h1>", rendered, rendered)
        self.assertNotIn("<h1>Shadowing variable = NOT SHADOWED</h1>", rendered, rendered)

    # NOTE: Second arg in tuple are expected values passed through components.
    @parametrize_context_behavior(
        [
            ("django", ("passed_in", "passed_in")),
            ("isolated", ("passed_in", "")),
        ]
    )
    def test_parent_args_available_in_slots(self, context_behavior_data):
        first_val, second_val = context_behavior_data

        template_str: dc.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'parent_with_args' parent_value='passed_in' %}
                {% fill 'content' %}
                    {% component name='variable_display' shadowing_variable='value_from_slot' new_variable=inner_parent_value %}
                    {% endcomponent %}
                {% endfill %}
            {% endcomponent %}
            """  # noqa: E501
        template = Template(template_str)
        rendered = template.render(Context())
        self.assertHTMLEqual(
            rendered,
            f"""
            <div>
                <h1>Parent content</h1>
                <h1>Shadowing variable = {first_val}</h1>
                <h1>Uniquely named variable = unique_val</h1>
            </div>
            <div>
                <h1>Shadowing variable = value_from_slot</h1>
                <h1>Uniquely named variable = {second_val}</h1>
            </div>
            """,
        )


class ContextCalledOnceTests(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        dc.registry.register(name="incrementer", component=IncrementerComponent)

    @parametrize_context_behavior(["django", "isolated"])
    def test_one_context_call_with_simple_component(self):
        template_str: dc.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component name='incrementer' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context()).strip().replace("\n", "")
        self.assertHTMLEqual(
            rendered,
            '<p class="incrementer">value=1;calls=1</p>',
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_one_context_call_with_simple_component_and_arg(self):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component name='incrementer' value='2' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context()).strip()

        self.assertHTMLEqual(rendered, '<p class="incrementer">value=3;calls=1</p>', rendered)

    @parametrize_context_behavior(["django", "isolated"])
    def test_one_context_call_with_component(self):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component 'incrementer' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context()).strip()

        self.assertHTMLEqual(rendered, '<p class="incrementer">value=1;calls=1</p>', rendered)

    @parametrize_context_behavior(["django", "isolated"])
    def test_one_context_call_with_component_and_arg(self):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component 'incrementer' value='3' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context()).strip()

        self.assertHTMLEqual(rendered, '<p class="incrementer">value=4;calls=1</p>', rendered)

    @parametrize_context_behavior(["django", "isolated"])
    def test_one_context_call_with_slot(self):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component 'incrementer' %}
                {% fill 'content' %}
                    <p>slot</p>
                {% endfill %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context()).strip()

        self.assertHTMLEqual(
            rendered,
            '<p class="incrementer">value=1;calls=1</p>\n<p>slot</p>',
            rendered,
        )

    @parametrize_context_behavior(["django", "isolated"])
    def test_one_context_call_with_slot_and_arg(self):
        template_str: dc.django_html = """
            {% load component_tags %}
            {% component 'incrementer' value='3' %}
                {% fill 'content' %}
                    <p>slot</p>
                {% endfill %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context()).strip()

        self.assertHTMLEqual(
            rendered,
            '<p class="incrementer">value=4;calls=1</p>\n<p>slot</p>',
            rendered,
        )


class ComponentsCanAccessOuterContext(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        dc.registry.register(name="simple_component", component=SimpleComponent)

    # NOTE: Second arg in tuple is expected value.
    @parametrize_context_behavior(
        [
            ("django", "outer_value"),
            ("isolated", ""),
        ]
    )
    def test_simple_component_can_use_outer_context(self, context_behavior_data):
        template_str: dc.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'simple_component' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({"variable": "outer_value"}))
        self.assertHTMLEqual(
            rendered,
            f"""
            Variable: <strong> {context_behavior_data} </strong>
            """,
        )


class IsolatedContextTests(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        dc.registry.register(name="simple_component", component=SimpleComponent)

    @parametrize_context_behavior(["django", "isolated"])
    def test_simple_component_can_pass_outer_context_in_args(self):
        template_str: dc.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'simple_component' variable only %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({"variable": "outer_value"})).strip()
        self.assertIn("outer_value", rendered, rendered)

    @parametrize_context_behavior(["django", "isolated"])
    def test_simple_component_cannot_use_outer_context(self):
        template_str: dc.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'simple_component' only %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({"variable": "outer_value"})).strip()
        self.assertNotIn("outer_value", rendered, rendered)


class IsolatedContextSettingTests(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        dc.registry.register(name="simple_component", component=SimpleComponent)

    @parametrize_context_behavior(["isolated"])
    def test_component_tag_includes_variable_with_isolated_context_from_settings(
        self,
    ):
        template_str: dc.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'simple_component' variable %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({"variable": "outer_value"}))
        self.assertIn("outer_value", rendered, rendered)

    @parametrize_context_behavior(["isolated"])
    def test_component_tag_excludes_variable_with_isolated_context_from_settings(
        self,
    ):
        template_str: dc.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'simple_component' %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({"variable": "outer_value"}))
        self.assertNotIn("outer_value", rendered, rendered)

    @parametrize_context_behavior(["isolated"])
    def test_component_includes_variable_with_isolated_context_from_settings(
        self,
    ):
        template_str: dc.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'simple_component' variable %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({"variable": "outer_value"}))
        self.assertIn("outer_value", rendered, rendered)

    @parametrize_context_behavior(["isolated"])
    def test_component_excludes_variable_with_isolated_context_from_settings(
        self,
    ):
        template_str: dc.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'simple_component' %}
            {% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({"variable": "outer_value"}))
        self.assertNotIn("outer_value", rendered, rendered)


class OuterContextPropertyTests(BaseTestCase):
    class OuterContextComponent(dc.Component):
        template: dc.django_html = """
            Variable: <strong>{{ variable }}</strong>
        """

        def get_context_data(self):
            return self.outer_context.flatten()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        dc.registry.register(name="outer_context_component", component=cls.OuterContextComponent)

    @parametrize_context_behavior(["django", "isolated"])
    def test_outer_context_property_with_component(self):
        template_str: dc.django_html = """
            {% load component_tags %}{% component_dependencies %}
            {% component 'outer_context_component' only %}{% endcomponent %}
        """
        template = Template(template_str)
        rendered = template.render(Context({"variable": "outer_value"})).strip()
        self.assertIn("outer_value", rendered, rendered)


class ContextVarsIsFilledTests(BaseTestCase):
    class IsFilledVarsComponent(dc.Component):
        template: dc.django_html = """
            {% load component_tags %}
            <div class="frontmatter-component">
                {% slot "title" default %}{% endslot %}
                {% slot "my_title" %}{% endslot %}
                {% slot "my title 1" %}{% endslot %}
                {% slot "my-title-2" %}{% endslot %}
                {% slot "escape this: #$%^*()" %}{% endslot %}
                {{ component_vars.is_filled|safe }}
            </div>
        """

    class ComponentWithConditionalSlots(dc.Component):
        template: dc.django_html = """
            {# Example from django-components/issues/98 #}
            {% load component_tags %}
            <div class="frontmatter-component">
                <div class="title">{% slot "title" %}Title{% endslot %}</div>
                {% if component_vars.is_filled.subtitle %}
                    <div class="subtitle">
                        {% slot "subtitle" %}Optional subtitle
                        {% endslot %}
                    </div>
                {% endif %}
            </div>
        """

    class ComponentWithComplexConditionalSlots(dc.Component):
        template: dc.django_html = """
            {# Example from django-components/issues/98 #}
            {% load component_tags %}
            <div class="frontmatter-component">
                <div class="title">{% slot "title" %}Title{% endslot %}</div>
                {% if component_vars.is_filled.subtitle %}
                    <div class="subtitle">{% slot "subtitle" %}Optional subtitle{% endslot %}</div>
                {% elif component_vars.is_filled.alt_subtitle %}
                    <div class="subtitle">{% slot "alt_subtitle" %}Why would you want this?{% endslot %}</div>
                {% else %}
                <div class="warning">Nothing filled!</div>
                {% endif %}
            </div>
        """

    class ComponentWithNegatedConditionalSlot(dc.Component):
        template: dc.django_html = """
            {# Example from django-components/issues/98 #}
            {% load component_tags %}
            <div class="frontmatter-component">
                <div class="title">{% slot "title" %}Title{% endslot %}</div>
                {% if not component_vars.is_filled.subtitle %}
                <div class="warning">Subtitle not filled!</div>
                {% else %}
                    <div class="subtitle">{% slot "alt_subtitle" %}Why would you want this?{% endslot %}</div>
                {% endif %}
            </div>
        """

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        dc.registry.register("is_filled_vars", cls.IsFilledVarsComponent)
        dc.registry.register("conditional_slots", cls.ComponentWithConditionalSlots)
        dc.registry.register(
            "complex_conditional_slots",
            cls.ComponentWithComplexConditionalSlots,
        )
        dc.registry.register("negated_conditional_slot", cls.ComponentWithNegatedConditionalSlot)

    @classmethod
    def tearDownClass(cls) -> None:
        super().tearDownClass()
        dc.registry.clear()

    @parametrize_context_behavior(["django", "isolated"])
    def test_is_filled_vars(self):
        template: dc.django_html = """
            {% load component_tags %}
            {% component "is_filled_vars" %}
                {% fill "title" %}{% endfill %}
                {% fill "my-title-2" %}{% endfill %}
                {% fill "escape this: #$%^*()" %}{% endfill %}
            {% endcomponent %}
        """
        rendered = Template(template).render(Context())
        expected = """
            <div class="frontmatter-component">
                {'title': True,
                'my_title': False,
                'my_title_1': False,
                'my_title_2': True,
                'escape_this_________': True}
            </div>
        """
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_is_filled_vars_default(self):
        template: dc.django_html = """
            {% load component_tags %}
            {% component "is_filled_vars" %}
                bla bla
            {% endcomponent %}
        """
        rendered = Template(template).render(Context())
        expected = """
            <div class="frontmatter-component">
                bla bla
                {'title': True,
                'my_title': False,
                'my_title_1': False,
                'my_title_2': False,
                'escape_this_________': False}
            </div>
        """
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_simple_component_with_conditional_slot(self):
        template: dc.django_html = """
            {% load component_tags %}
            {% component "conditional_slots" %}{% endcomponent %}
        """
        expected = """
            <div class="frontmatter-component">
            <div class="title">
            Title
            </div>
            </div>
        """
        rendered = Template(template).render(Context({}))
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_component_with_filled_conditional_slot(self):
        template: dc.django_html = """
            {% load component_tags %}
            {% component "conditional_slots" %}
            {% fill "subtitle" %} My subtitle {% endfill %}
            {% endcomponent %}
        """
        expected = """
            <div class="frontmatter-component">
                <div class="title">
                    Title
                </div>
                <div class="subtitle">
                    My subtitle
                </div>
            </div>
        """
        rendered = Template(template).render(Context({}))
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_elif_of_complex_conditional_slots(self):
        template: dc.django_html = """
            {% load component_tags %}
            {% component "complex_conditional_slots" %}
                {% fill "alt_subtitle" %} A different subtitle {% endfill %}
            {% endcomponent %}
        """
        expected = """
           <div class="frontmatter-component">
             <div class="title">
                Title
             </div>
             <div class="subtitle">
                A different subtitle
             </div>
           </div>
        """
        rendered = Template(template).render(Context({}))
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_else_of_complex_conditional_slots(self):
        template: dc.django_html = """
           {% load component_tags %}
           {% component "complex_conditional_slots" %}
           {% endcomponent %}
        """
        expected = """
           <div class="frontmatter-component">
             <div class="title">
             Title
             </div>
            <div class="warning">Nothing filled!</div>
           </div>
        """
        rendered = Template(template).render(Context({}))
        self.assertHTMLEqual(rendered, expected)

    @parametrize_context_behavior(["django", "isolated"])
    def test_component_with_negated_conditional_slot(self):
        template: dc.django_html = """
            {% load component_tags %}
            {% component "negated_conditional_slot" %}
                {# Whoops! Forgot to fill a slot! #}
            {% endcomponent %}
        """
        expected = """
            <div class="frontmatter-component">
                <div class="title">
                Title
                </div>
                <div class="warning">Subtitle not filled!</div>
            </div>
        """
        rendered = Template(template).render(Context({}))
        self.assertHTMLEqual(rendered, expected)
